import json
import os
from dataclasses import dataclass
from test.examples.stages.cache_stages import (
    filerefcacher_stage,
    filerefcacher_stage_multifile,
)

import numpy as np
import pandas as pd
import pytest

import curifactory as cf
from curifactory.caching import (
    Cacheable,
    JsonCacher,
    PandasCacher,
    PandasCsvCacher,
    PandasJsonCacher,
    PickleCacher,
    RawJupyterNotebookCacher,
    _PandasIOType,
)
from curifactory.reporting import JsonReporter


# TODO: necessary? configured_test_manager already does this
@pytest.fixture()
def clear_stage_run(configured_test_manager):
    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    try:
        os.remove(ran_path)
    except FileNotFoundError:
        pass
    yield
    try:
        os.remove(ran_path)
    except FileNotFoundError:
        pass


def test_filerefcacher_stores_multiple_paths(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should correctly store a list of files in the saved json."""
    r = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage_multifile(r)

    argshash = r.params.hash
    expected_list = [
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_multifile_my_files/thing{i}",
        )
        for i in range(5)
    ]

    with open(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_multifile_output_paths.json",
        ),
    ) as infile:
        filelist = json.load(infile)
        assert filelist == expected_list

        for filename in filelist:
            assert os.path.exists(filename)


def test_filerefcacher_stores_single_path(configured_test_manager, clear_stage_run):
    r = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage(r)

    argshash = r.params.hash
    expected_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{argshash}_filerefcacher_stage_my_file",
    )

    with open(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{argshash}_filerefcacher_stage_output_path.json",
        ),
    ) as infile:
        filelist = json.load(infile)
        assert filelist == expected_path


def test_filerefcacher_shortcircuits(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should short-circuit if all files in the filelist already exist."""
    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage_multifile(r0)

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    assert os.path.exists(ran_path)
    os.remove(ran_path)

    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage_multifile(r1)
    assert not os.path.exists(ran_path)


def test_filerefcacher_runs_when_file_missing(configured_test_manager, clear_stage_run):
    """FileReferenceCacher should _not_ short-circuit if any of the files in the filelist are missing."""
    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage_multifile(r0)

    ran_path = os.path.join(configured_test_manager.cache_path, "stage_ran")
    assert os.path.exists(ran_path)
    os.remove(ran_path)
    os.remove(
        os.path.join(
            configured_test_manager.cache_path,
            f"test_{r0.params.hash}_filerefcacher_stage_multifile_my_files/thing1",
        )
    )

    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    filerefcacher_stage_multifile(r1)
    assert os.path.exists(ran_path)


def test_reportables_are_cached(configured_test_manager):
    """Running a stage with a reportable should cache the reportable and a list of reportable cache files."""

    @cf.stage(None, ["test_output"], [PickleCacher])
    def basic_reportable(record):
        record.report(JsonReporter({"test": "hello world"}))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r0)

    list_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_basic_reportable_reportables_file_list.json",
    )

    reportable_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_basic_reportable_reportables/test_basic_reportable_0.pkl",
    )

    assert os.path.exists(list_path)

    with open(list_path) as infile:
        paths = json.load(infile)

    assert len(paths) == 1
    assert paths[0] == reportable_path
    assert os.path.exists(reportable_path)


def test_named_reportables_are_cached(configured_test_manager):
    """Running a stage with a named reportable should cache the reportable and a list of reportable cache files."""

    @cf.stage(None, ["test_output"], [PickleCacher])
    def basic_reportable(record):
        record.report(JsonReporter({"test": "hello world"}, name="thing"))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r0)

    list_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_basic_reportable_reportables_file_list.json",
    )

    reportable_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_basic_reportable_reportables/test_basic_reportable_thing.pkl",
    )

    assert os.path.exists(list_path)

    with open(list_path) as infile:
        paths = json.load(infile)

    assert len(paths) == 1
    assert paths[0] == reportable_path
    assert os.path.exists(reportable_path)


def test_cached_reportables_loaded_without_doubling_name(configured_test_manager):
    """Re-loading cached reportables should not double name!

    (This was caused previously by re-reporting on the record, which prefixed the
    name in-place.)
    """
    run_count = 0

    @cf.stage(None, ["test_output"], [PickleCacher])
    def basic_reportable(record):
        nonlocal run_count
        run_count += 1
        record.report(JsonReporter({"test": "hello world"}))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r0)

    assert len(configured_test_manager.reportables) == 1

    # run again in a new record with exact same config, so it will find cached
    # things.
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r1)

    # make sure we didn't run the stage, so we actually had to load reportables
    assert run_count == 1

    assert len(configured_test_manager.reportables) == 2
    assert configured_test_manager.reportables[0].name is None
    assert configured_test_manager.reportables[1].name is None
    assert (
        configured_test_manager.reportables[0].qualified_name
        == "test_basic_reportable_0"
    )
    assert (
        configured_test_manager.reportables[1].qualified_name
        == "test_basic_reportable_1"
    )


def test_cached_named_reportables_loaded_without_doubling_name(configured_test_manager):
    """Re-loading cached named reportables should not double name!

    (This was caused previously by re-reporting on the record, which prefixed the
    name in-place.)
    """
    run_count = 0

    @cf.stage(None, ["test_output"], [PickleCacher])
    def basic_reportable(record):
        nonlocal run_count
        run_count += 1
        record.report(JsonReporter({"test": "hello world"}, name="thing"))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r0)

    assert len(configured_test_manager.reportables) == 1

    # run again in a new record with exact same config, so it will find cached
    # things.
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    basic_reportable(r1)

    # make sure we didn't run the stage, so we actually had to load reportables
    assert run_count == 1

    assert len(configured_test_manager.reportables) == 2
    assert (
        configured_test_manager.reportables[0].name
        == configured_test_manager.reportables[1].name
    )
    assert (
        configured_test_manager.reportables[0].qualified_name
        == configured_test_manager.reportables[1].qualified_name
    )


def test_aggregate_reportables_are_cached(configured_test_manager):
    """Running an aggregate stage with a reportable should cache the reportable and a list of reportable cache files.
    (using the record's aggregate combo hash)"""

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def basic_agg_reportable(record, records):
        record.report(JsonReporter({"test": "hello world"}))
        return "test"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    print("hash", r0.params.hash)
    basic_agg_reportable(r0)
    print("hash", r0.params.hash)
    assert r0.params.hash is not None

    list_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.combo_hash}_basic_agg_reportable_reportables_file_list.json",
    )

    reportable_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.combo_hash}_basic_agg_reportable_reportables/(Aggregate)_test_basic_agg_reportable_0.pkl",
    )

    print([thing for thing in os.listdir(configured_test_manager.cache_path)])
    print("---")
    print(list_path)
    print(reportable_path)

    assert os.path.exists(list_path)

    with open(list_path) as infile:
        paths = json.load(infile)

    assert len(paths) == 1
    print("PATHS:", paths)
    assert paths[0] == reportable_path
    assert os.path.exists(reportable_path)


# TODO: test that those reportables are correctly reloaded when re-run


# def test_reportables_are_cached_with_store_full(configured_test_manager):
#     pass


def test_aggregate_args_no_records_loads_cache(configured_test_manager):
    """Calling an aggregate stage with valid args, twice, should load from cache and not execute."""
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r0 = test_agg(r0, [])
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = test_agg(r1, [])

    assert call_count == 1
    assert r1.state["test_output"] == "hello world!"


def test_aggregate_args_records_loads_cache(configured_test_manager):
    """Calling an aggregate stage with valid args, twice, with other non-overwrite records involved
    should load from cache and not execute."""
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    rA = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testA"))
    rB = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testB"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r0 = test_agg(r0, [rA, rB])
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = test_agg(r1, [rA, rB])

    assert call_count == 1
    assert r1.state["test_output"] == "hello world!"


def test_aggregate_args_overwrite_no_records_doesnot_load_cache(
    configured_test_manager,
):
    """Calling an aggregate stage with valid args with overwrite specified, twice, should NOT load
    from cache and execute."""
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    r0 = cf.Record(
        configured_test_manager, cf.ExperimentParameters(name="test", overwrite=True)
    )
    r0 = test_agg(r0, [])
    r1 = cf.Record(
        configured_test_manager, cf.ExperimentParameters(name="test", overwrite=True)
    )
    r1 = test_agg(r1, [])

    assert call_count == 2


def test_aggregate_args_records_overwrite_loads_cache(configured_test_manager):
    """Calling an aggregate stage with valid args, twice, with other records with overwrite
    should still load from cache and not execute.

    This is a weird edge case, running with the experiment CLI should ensure this doesn't happen
    (since --overwrite will apply to all args, it's set on an artifactmanager-level.) This could in
    principle only happen in a live context (like a notebook).

    The expected behavior, given that we have no intelligent-DAG-based-overwrites yet, is to use the
    cache anyway. The only reason we check the input records in other cases is when the agg stage record
    has no arguments of its own and so overwrite must be inferred. In this case however, we have an actual
    set of arguments to use, so there's no way to distinguish between whether an overwrite of False is
    "default" or intentional, so we just assume we trust that the args are correct.
    """
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    rA = cf.Record(
        configured_test_manager, cf.ExperimentParameters(name="testA", overwrite=True)
    )
    rB = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testB"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r0 = test_agg(r0, [rA, rB])
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = test_agg(r1, [rA, rB])

    assert call_count == 1


def test_aggregate_no_args_no_records_doesnot_load_cache(configured_test_manager):
    """Calling an aggregate stage with no args, twice, should NOT load from cache and execute.

    This is a relatively pointless situation, but just to cover the bases and establish expected functionality,
    a no args situation should not use the cache when no other records present, as we have no reasonable
    combo hash and no information about whether overwrite is specified on anything else or not.
    """
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    r0 = cf.Record(configured_test_manager, None)
    r0 = test_agg(r0, [])
    r1 = cf.Record(configured_test_manager, None)
    r1 = test_agg(r1, [])

    assert call_count == 2


def test_aggregate_no_args_records_loads_cache(configured_test_manager):
    """Calling an aggregate stage with no args, twice, with other records should load from cache and not execute.

    Even though there's no args, we can establish a re-execution based on the combohash of the passed records, so
    allow cache usage.
    """
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    rA = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testA"))
    rB = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testB"))

    r0 = cf.Record(configured_test_manager, None)
    r0 = test_agg(r0, [rA, rB])
    r1 = cf.Record(configured_test_manager, None)
    r1 = test_agg(r1, [rA, rB])

    assert call_count == 1
    assert r1.state["test_output"] == "hello world!"


# TODO: (05/13/2022) what about stages after such an agg? Without global overwrite specified, they won't overwrite, since
# it's not part of an args. This is the same problem as giving an --overwrite-stage though, so unsure that this needs to be
# handled. The below test case can really only apply if you're manually doing weird things outside of experiment CLI
def test_aggregate_no_args_records_overwrite_doesnot_load_cache(
    configured_test_manager,
):
    """Calling an aggregate stage with no args, twice, with other records with overwrite should NOT load from
    cache and execute.

    If at least one of the associated records has overwrite specified, that should carry through into
    """
    call_count = 0

    @cf.aggregate(None, ["test_output"], [PickleCacher])
    def test_agg(record, records):
        nonlocal call_count
        call_count += 1
        return "hello world!"

    rA = cf.Record(
        configured_test_manager, cf.ExperimentParameters(name="testA", overwrite=True)
    )
    rB = cf.Record(configured_test_manager, cf.ExperimentParameters(name="testB"))

    r0 = cf.Record(configured_test_manager, None)
    r0 = test_agg(r0, [rA, rB])
    r1 = cf.Record(configured_test_manager, None)
    r1 = test_agg(r1, [rA, rB])

    assert call_count == 2


def test_get_path_file_included_in_full_store(configured_test_manager):
    """A file manually saved within a stage using get_path should correctly be
    copied to the run folder in a full-store run."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["other_output"], [PickleCacher])
    def custom_output(record):
        path = record.get_path("my_extra_file.txt")
        with open(path, "w") as outfile:
            outfile.write("Hello world!")

        return 13

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    custom_output(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    regular_custom_output_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_custom_output_my_extra_file.txt",
    )
    full_store_custom_output_path = os.path.join(
        full_store_path, f"test_{r0.params.hash}_custom_output_my_extra_file.txt"
    )
    assert os.path.exists(regular_custom_output_path)
    assert os.path.exists(full_store_custom_output_path)


def test_get_dir_folder_included_in_full_store(configured_test_manager):
    """File(s) manually saved within a stage using get_dir should correctly be
    copied to the run folder in a full-store run."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["other_output"], [PickleCacher])
    def custom_output(record):
        path = record.get_dir("my_extra_dir")
        with open(f"{path}/testfile.txt", "w") as outfile:
            outfile.write("Hello world!")

        return 13

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    custom_output(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    regular_custom_output_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_custom_output_my_extra_dir/" "testfile.txt",
    )
    full_store_custom_output_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_custom_output_my_extra_dir/" "testfile.txt",
    )
    assert os.path.exists(regular_custom_output_path)
    assert os.path.exists(full_store_custom_output_path)


def test_get_path_file_excluded_in_full_store_when_not_tracked(configured_test_manager):
    """A file manually saved within a stage using get_path should NOT be
    copied to the run folder in a full-store run when not tracked."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["other_output"], [PickleCacher])
    def custom_output(record):
        path = record.get_path("my_extra_file.txt", track=False)
        with open(path, "w") as outfile:
            outfile.write("Hello world!")

        return 13

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    custom_output(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    regular_custom_output_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_custom_output_my_extra_file.txt",
    )
    full_store_custom_output_path = os.path.join(
        full_store_path, f"test_{r0.params.hash}_custom_output_my_extra_file.txt"
    )
    assert os.path.exists(regular_custom_output_path)
    assert not os.path.exists(full_store_custom_output_path)


def test_get_dir_folder_excluded_in_full_store_when_not_tracked(
    configured_test_manager,
):
    """File(s) manually saved within a stage using get_dir should NOT be
    copied to the run folder in a full-store run when not tracked."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["other_output"], [PickleCacher])
    def custom_output(record):
        path = record.get_dir("my_extra_dir", track=False)
        with open(f"{path}/testfile.txt", "w") as outfile:
            outfile.write("Hello world!")

        return 13

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    custom_output(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    regular_custom_output_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_custom_output_my_extra_dir/" "testfile.txt",
    )
    full_store_custom_output_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_custom_output_my_extra_dir/" "testfile.txt",
    )
    assert os.path.exists(regular_custom_output_path)
    assert not os.path.exists(full_store_custom_output_path)


def test_pandas_csv_cacher_with_df_with_comma(configured_test_manager):
    """The PandasCSVCacher shouldn't fail when given a dataframe containing a comma."""

    @cf.stage(None, ["output"], [PandasCsvCacher])
    def save_comma_df(record):
        data = {
            "col1": ["things, with commas"],
            "col2": ['other "possible breaking," things'],
        }
        df = pd.DataFrame(data=data)
        return df

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    save_comma_df(r0)

    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    save_comma_df(r1)

    df1 = r0.state["output"]
    df2 = r1.state["output"]

    assert list(df1.columns) == list(df2.columns)
    assert df1.to_dict() == df2.to_dict()


def test_pandas_json_cacher_with_df_no_recursion_error(configured_test_manager):
    """Regression test for weird bug in #3, caching a df with 17+ cols and 5+
    rows with the pandasjsoncacher shouldn't crash with a maximum recursion
    level reached."""

    data = np.random.rand(18, 6)

    @cf.stage(None, ["output"], [PandasJsonCacher])
    def save_large_df(record):
        df = pd.DataFrame(data)
        return df

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    save_large_df(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_save_large_df_output.json",
    )

    assert os.path.exists(path)
    df = pd.read_json(path)
    np.testing.assert_almost_equal(df.values, data)


@pytest.mark.parametrize(
    "io_format", ["csv", "json", "parquet", "pickle", "orc", "hdf5", "excel", "xml"]
)
def test_pandas_cacher_for_all_io_formats(configured_test_manager, io_format):
    """The PandasCacher should work for save and load for all IO formats."""

    io_format_enum = _PandasIOType[io_format]
    data = {
        "col1": ["things, with commas", "more rows", "MOAR!"],
        "col2": ['other "possible breaking," things', "Tres Commas", "No"],
        "col3": [42, -1, 3.14159],
    }
    saved_df = pd.DataFrame(data)

    @cf.stage(outputs=["df"], cachers=[PandasCacher(format=io_format)])
    def write_pandas(record):
        return saved_df

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    write_pandas(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_write_pandas_df",
    )
    assert os.path.exists(f"{path}.{io_format_enum.value}")

    loaded_df = PandasCacher(f"{path}.{io_format_enum.value}", format=io_format).load()

    assert saved_df.equals(loaded_df)


@pytest.mark.parametrize(
    # fmt: off
    "io_format,to_args,read_args",
    [
        ("csv", {}, {"index_col": 0, "dtype_backend": "pyarrow"}),
        ("json", {"double_precision": 15}, {"dtype_backend": "pyarrow"}),
        ("parquet", {}, {"dtype_backend": "pyarrow"}),
        ("pickle", {}, {}),
        ("orc", {}, {"dtype_backend": "pyarrow"}),
        ("excel", {}, {"index_col": 0, "dtype_backend": "pyarrow"}),
        ("xml", {"index": False}, {"dtype_backend": "pyarrow"}),
    ]
    # fmt: on
)
def test_pandas_cacher_for_all_io_formats_with_args(
    configured_test_manager, io_format, to_args, read_args
):
    """The PandasCacher should work for save and load for all IO formats
    with input to_args and read_args."""

    io_format_enum = _PandasIOType[io_format]
    data = {
        "col1": ["things, with commas", "more rows", "MOAR!"],
        "col2": ['other "possible breaking," things', "Tres Commas", "No"],
        "col3": [42, -1, 3.14159],
    }
    saved_df = pd.DataFrame(data)
    saved_df = saved_df.convert_dtypes(dtype_backend="pyarrow")
    if io_format_enum in (_PandasIOType.csv, _PandasIOType.excel):
        saved_df.index = saved_df.index.astype("int64[pyarrow]", copy=False)

    @cf.stage(
        outputs=["df"],
        cachers=[PandasCacher(format=io_format, to_args=to_args, read_args=read_args)],
    )
    def write_pandas(record):
        return saved_df

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    write_pandas(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_write_pandas_df",
    )
    assert os.path.exists(f"{path}.{io_format_enum.value}")

    loaded_df = PandasCacher(
        f"{path}.{io_format_enum.value}",
        format=io_format,
        to_args=to_args,
        read_args=read_args,
    ).load()

    assert saved_df.equals(loaded_df)


@pytest.mark.parametrize(
    # fmt: off
    "cacher_args,suffix,expected_path",
    [
        (dict(name="test"), None, "test_hash_test_stage_test.pkl"),
        (dict(name="test"), "_metadata.json", "test_hash_test_stage_test_metadata.json"),
        (dict(name="test"), "_thing", "test_hash_test_stage_test_thing"),
        (dict(name="test.pkl"), None, "test_hash_test_stage_test.pkl"),
        (dict(path_override="test/examples/data/cache/test"), None, "test"),
        (dict(path_override="test/examples/data/cache/test.pkl"), None, "test.pkl"),
        (dict(path_override="test/examples/data/cache/test"), "_metadata.json", "test_metadata.json"),
        (dict(path_override="test/examples/data/cache/test.pkl"), "_metadata.json", "test_metadata.json"),
        (dict(name="test", prefix="someprefix"), None, "someprefix_hash_test_stage_test.pkl"),
        (dict(name="test", prefix="someprefix"), "_metadata.json", "someprefix_hash_test_stage_test_metadata.json"),
        (dict(name="test", subdir="01_raw"), None, "01_raw/test_hash_test_stage_test.pkl"),
        (dict(name="test", subdir="01_raw"), "_metadata.json", "01_raw/test_hash_test_stage_test_metadata.json"),
        (dict(name="test", subdir="01_raw", prefix="someprefix"), None, "01_raw/someprefix_hash_test_stage_test.pkl"),
        (dict(name="test", subdir="01_raw", prefix="someprefix"), "_metadata.json", "01_raw/someprefix_hash_test_stage_test_metadata.json"),
    ]
    # fmt: on
)
def test_cacheable_get_path(
    configured_test_manager, cacher_args, suffix, expected_path
):
    """A cacher's get_path should correctly handle different combinations of path inputs and
    requested suffixes."""
    configured_test_manager.current_stage_name = "test_stage"
    r = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r.params.hash = "hash"

    cacher = PickleCacher(**cacher_args, record=r)

    path_prefix = "test/examples/data/cache/"
    path = cacher.get_path(suffix)
    assert path == path_prefix + expected_path


def test_cacher_outputs_metadata(configured_test_manager):
    """A basic cacher output should also output a metadata file associated with it."""

    @cf.stage(None, ["output"], [PickleCacher])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_output_thing_output.pkl",
    )
    metadata_path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_output_thing_output_metadata.json",
    )

    assert os.path.exists(path)
    assert os.path.exists(metadata_path)


def test_cacher_outputs_metadata_storefull(configured_test_manager):
    """A cacher output should copy the associated metadata file to the full store if full store mode."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["output"], [PickleCacher])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    metadata_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_output_thing_output_metadata.json",
    )
    assert os.path.exists(metadata_path)


def test_manual_static_path_cacher_outputs_metadata(configured_test_manager):
    """Metadata should still output at the correct path when using a manual static-path cacher."""

    @cf.stage()
    def output_thing(record):
        cacher = PickleCacher(
            "test/examples/data/cache/raw_path_file.pkl", record=record
        )
        cacher.save("hello world")
        cacher.save_metadata()

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    path = "test/examples/data/cache/raw_path_file.pkl"
    loader = PickleCacher(path)
    metadata = loader.load_metadata()

    assert os.path.exists(path)
    assert os.path.exists("test/examples/data/cache/raw_path_file_metadata.json")
    assert loader.load() == "hello world"
    assert metadata["stage"] == "output_thing"


def test_load_metadata_with_manual_cacher_from_stage_cacher_path(
    configured_test_manager,
):
    """Metadata for a static-path cacher created based on a stage cacher's path should still be loadable."""

    @cf.stage(None, ["output"], [PickleCacher])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_output_thing_output.pkl",
    )

    manual_cacher = PickleCacher(path)
    metadata = manual_cacher.load_metadata()

    assert metadata is not None
    assert metadata["artifact_name"] == "output"


def test_overwrites_stage_doesnot_break_storefull_of_other_stages(
    configured_test_manager, alternate_test_manager2
):
    """An experiment run twice, the second time with an overwrite-stage flag and store-full, should
    transfer both the newly overwritten artifacts into the run folder as well as the old non-overwritten
    artifacts from the other stages."""
    configured_test_manager.store_full = True
    alternate_test_manager2.store_full = True
    alternate_test_manager2.overwrite_stages = ["overwritten_output"]

    nonoverwritten_call_count = 0
    overwritten_call_count = 0

    @cf.stage(None, ["firstoutput"], [PickleCacher(prefix="common")])
    def non_overwritten_output(record):
        nonlocal nonoverwritten_call_count
        nonoverwritten_call_count += 1
        return "hello!"

    @cf.stage(["firstoutput"], ["secondoutput"], [PickleCacher(prefix="common")])
    def overwritten_output(record, firstoutput):
        nonlocal overwritten_call_count
        overwritten_call_count += 1
        return "world!"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = cf.Record(alternate_test_manager2, cf.ExperimentParameters(name="test2"))

    overwritten_output(non_overwritten_output(r0))
    overwritten_output(non_overwritten_output(r1))

    assert nonoverwritten_call_count == 1
    assert overwritten_call_count == 2

    second_full_store_path = f"{alternate_test_manager2.runs_path}/test2_1_{alternate_test_manager2.get_str_timestamp()}/artifacts"
    nonoverwritten_output_path = os.path.join(
        second_full_store_path,
        f"common_{r1.params.hash}_non_overwritten_output_firstoutput_metadata.json",
    )
    overwritten_output_path = os.path.join(
        second_full_store_path,
        f"common_{r1.params.hash}_overwritten_output_secondoutput_metadata.json",
    )

    assert os.path.exists(nonoverwritten_output_path)
    assert os.path.exists(overwritten_output_path)


def test_non_tracked_cacher_does_not_copy_metadata_to_full_store(
    configured_test_manager,
):
    """A non-tracked cacher should not copy the metadata file to the full store."""
    configured_test_manager.store_full = True

    @cf.stage(None, ["output"], [PickleCacher(track=False)])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    metadata_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_output_thing_output_metadata.json",
    )
    assert not os.path.exists(metadata_path)


def test_no_metadata_written_to_dry_cache_folder(
    configured_test_manager,
):
    """Metadata should not be written out to a dry-cache cache folder."""
    configured_test_manager.dry_cache = True

    @cf.stage(None, ["output"], [PickleCacher])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"

    metadata_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_output_thing_output_metadata.json",
    )
    assert not os.path.exists(metadata_path)


def test_existing_metadata_not_overwritten_when_cache_used(
    configured_test_manager, alternate_test_manager2
):
    """When cached values loaded in, existing metadata file should not be overwritten. Tested here
    with cross-caching"""

    @cf.stage(None, ["output"], [PickleCacher(prefix="commondata")])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    r1 = cf.Record(alternate_test_manager2, cf.ExperimentParameters(name="test"))
    output_thing(r1)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"commondata_{r1.params.hash}_output_thing_output_metadata.json",
    )
    metadata = JsonCacher(path).load()
    assert (
        metadata["manager_run_info"]["experiment_name"] == "test"
    )  # rather than test2


def test_uses_existing_metadata_in_full_store_when_cache_used(
    configured_test_manager, alternate_test_manager2
):
    """A full store output that's using a cached value should transfer the _existing_ metadata file
    to the full store."""
    configured_test_manager.store_full = True
    alternate_test_manager2.store_full = True

    @cf.stage(None, ["output"], [PickleCacher(prefix="commondata")])
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    r1 = cf.Record(alternate_test_manager2, cf.ExperimentParameters(name="test"))
    output_thing(r1)

    full_store_path = f"{alternate_test_manager2.runs_path}/test2_1_{alternate_test_manager2.get_str_timestamp()}/artifacts"
    path = os.path.join(
        full_store_path,
        f"commondata_{r1.params.hash}_output_thing_output_metadata.json",
    )
    metadata = JsonCacher(path).load()
    assert (
        metadata["manager_run_info"]["experiment_name"] == "test"
    )  # rather than test2


def test_cacher_getpath_keeps_stagename_after_later_stages(configured_test_manager):
    """A cacher created in one stage should still return the same get_paths even after later
    stages have executed."""
    cacher = PickleCacher()

    @cf.stage(None, ["output"], [cacher])
    def output_thing(record):
        return "Hello world"

    @cf.stage(["output"], ["output2"])
    def do_something_else(record, output):
        return output + ", kthxbye"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r0 = output_thing(r0)
    assert cacher.stage == "output_thing"
    first_path = cacher.get_path()

    do_something_else(r0)
    assert cacher.stage == "output_thing"
    second_path = cacher.get_path()

    assert first_path == second_path


def test_path_override_cacher_saves_to_that_path(configured_test_manager):
    """A cacher with path-override set should save the output to that path."""

    @cf.stage(
        None, ["output"], [PickleCacher("test/examples/data/cache/raw_path_file.pkl")]
    )
    def output_thing(record):
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    path = "test/examples/data/cache/raw_path_file.pkl"
    loader = PickleCacher(path)
    metadata = loader.load_metadata()

    assert os.path.exists(path)
    assert os.path.exists("test/examples/data/cache/raw_path_file_metadata.json")
    assert loader.load() == "Hello world"
    assert metadata["stage"] == "output_thing"


def test_separate_managers_no_crosscache_by_default(
    configured_test_manager, alternate_test_manager2
):
    """Two separate managers with different paths and a common stage withOUT custom prefix should
    _not_ use the same cached value."""
    run_count = 0

    @cf.stage(None, ["output"], [PickleCacher])
    def output_thing(record):
        nonlocal run_count
        run_count += 1
        return "Hello world"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = cf.Record(alternate_test_manager2, cf.ExperimentParameters(name="test"))
    output_thing(r0)
    output_thing(r1)
    assert run_count == 2


def test_separate_managers_common_prefix_cacher_crosscaches(
    configured_test_manager, alternate_test_manager2
):
    """Two separate managers with different paths but a common stage with custom prefix should
    both use the same cached value if the args are the same."""
    run_count = 0

    @dataclass
    class MyArgs(cf.ExperimentParameters):
        a: int = 5

    @cf.stage(None, ["output"], [PickleCacher(prefix="commondata")])
    def output_thing(record):
        nonlocal run_count
        run_count += 1
        return "Hello world"

    r0 = cf.Record(configured_test_manager, MyArgs(name="test", a=4))
    r1 = cf.Record(alternate_test_manager2, MyArgs(name="test2", a=4))
    output_thing(r0)
    output_thing(r1)
    assert run_count == 1


def test_separate_managers_common_prefix_cacher_no_crosscache_if_args_diff(
    configured_test_manager, alternate_test_manager2
):
    """Two separate managers with different paths but a common stage with custom prefix should
    _not_ use the same cached value if the args are not the same."""
    run_count = 0

    @dataclass
    class MyArgs(cf.ExperimentParameters):
        a: int = 5

    @cf.stage(None, ["output"], [PickleCacher(prefix="commondata")])
    def output_thing(record):
        nonlocal run_count
        run_count += 1
        return "Hello world"

    r0 = cf.Record(configured_test_manager, MyArgs(name="test", a=4))
    r1 = cf.Record(alternate_test_manager2, MyArgs(name="test2", a=5))
    output_thing(r0)
    output_thing(r1)
    assert run_count == 2


def test_cacher_with_record_get_path(configured_test_manager):
    """A manual cacher with record.get_path static path should correctly save and load to that path."""
    cacher = None

    @cf.stage()
    def manual_output_thing(record):
        nonlocal cacher
        cacher = JsonCacher(record.get_path("manualtest.json"))
        cacher.save(dict(message="hello world"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    manual_output_thing(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_manual_output_thing_manualtest.json",
    )
    assert os.path.exists(path)
    assert cacher.load()["message"] == "hello world"


def test_cacher_with_record_get_path_no_extension(configured_test_manager):
    """A manual cacher with record.get_path static path should correctly save and load to that path,
    even when an extension hasn't been implicitly provided in the record.get_path call
    """
    cacher = None

    @cf.stage()
    def manual_output_thing(record):
        nonlocal cacher
        cacher = JsonCacher(record.get_path("manualtest"))
        cacher.save(dict(message="hello world"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    manual_output_thing(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_manual_output_thing_manualtest",
    )
    assert os.path.exists(path)
    assert cacher.load()["message"] == "hello world"


def test_cacher_with_record_get_path_full_store(configured_test_manager):
    """A manual cacher with record.get_path static path should correctly store to that path in
    full store in full store mode."""
    configured_test_manager.store_full = True

    @cf.stage()
    def manual_output_thing(record):
        cacher = JsonCacher(record.get_path("manualtest.json"))
        cacher.save(dict(message="hello world"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    manual_output_thing(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"
    path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_manual_output_thing_manualtest.json",
    )
    assert os.path.exists(path)


def test_cacher_with_record_get_path_no_extension_full_store(configured_test_manager):
    """A manual cacher with record.get_path static path should correctly store to that path in
    full store in full store mode, even when an extension hasn't been implicitly provided to the
    record.get_path"""
    configured_test_manager.store_full = True
    cacher = None

    @cf.stage()
    def manual_output_thing(record):
        nonlocal cacher
        cacher = JsonCacher(record.get_path("manualtest"))
        cacher.save(dict(message="hello world"))

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    manual_output_thing(r0)

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"
    path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_manual_output_thing_manualtest",
    )
    assert os.path.exists(path)


def test_custom_cacher_using_get_dir_store_full(configured_test_manager):
    """A custom cacheable that uses the cacher's get_dir function should correctly create
    the directory and allow storing in it, and transfer it to a full store."""
    configured_test_manager.store_full = True

    class MultiCacher(Cacheable):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def save(self, obj):
            dirpath = self.get_dir()
            with open(f"{dirpath}/thing.json", "w") as outfile:
                json.dump(obj, outfile)
            with open(f"{dirpath}/what.txt", "w") as outfile:
                outfile.write("This is the confirmation that thing.json was written.")

        def load(self):
            dirpath = self.get_dir()
            with open(f"{dirpath}/thing.json") as infile:
                return json.load(infile)

    cacher = MultiCacher()

    @cf.stage(outputs=["output"], cachers=[cacher])
    def output_thing(record):
        return dict(message="hello world!")

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    output_thing(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_output_thing_output",
    )
    assert os.path.exists(path)
    assert os.path.exists(f"{path}/thing.json")
    assert os.path.exists(f"{path}/what.txt")

    full_store_path = f"{configured_test_manager.runs_path}/test_1_{configured_test_manager.get_str_timestamp()}/artifacts"
    full_path = os.path.join(
        full_store_path,
        f"test_{r0.params.hash}_output_thing_output",
    )
    assert os.path.exists(full_path)
    assert os.path.exists(f"{full_path}/thing.json")
    assert os.path.exists(f"{full_path}/what.txt")

    assert cacher.load()["message"] == "hello world!"


def test_raw_jupyter_notebook_cacher_saves_outputs(configured_test_manager):
    """The RawJuypterNotebookCacher should store both the cell content and
    the notebook file."""

    @cf.stage(outputs=["notebook"], cachers=[RawJupyterNotebookCacher])
    def write_notebook(record):
        cells = [
            ["print('hello world!')", "print('this is a test!')"],
            ["# this is another cell", "..."],
        ]
        return cells

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    write_notebook(r0)

    path = os.path.join(
        configured_test_manager.cache_path,
        f"test_{r0.params.hash}_write_notebook_notebook",
    )
    assert os.path.exists(f"{path}.ipynb")
    assert os.path.exists(f"{path}_cells.json")
    assert not os.path.exists(f"{path}.py")

    contents = JsonCacher(f"{path}.ipynb").load()
    assert len(contents["cells"]) == 2
    assert contents["cells"][0]["source"] == [
        "print('hello world!')\n",
        "print('this is a test!')",
    ]
    assert contents["cells"][1]["source"] == ["# this is another cell\n", "..."]


def test_extra_metadata_loaded_before_load(configured_test_manager):
    """Any extra metadata added in a cacher's save function should be available in the load function."""
    run_count = 0

    class UsesExtraMetadataCacher(Cacheable):
        def save(self, obj):
            self.extra_metadata["best_number"] = 13
            self.save_metadata()
            JsonCacher(self.get_path()).save(obj)

        def load(self):
            self.load_metadata()
            assert self.extra_metadata["best_number"] == 13
            return JsonCacher(self.get_path()).load()

    @cf.stage(outputs=["something"], cachers=[UsesExtraMetadataCacher])
    def do_thing(record):
        nonlocal run_count
        run_count += 1
        return "hello"

    r0 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test"))
    r1 = cf.Record(configured_test_manager, cf.ExperimentParameters(name="test2"))
    do_thing(r0)
    do_thing(r1)

    assert r0.state["something"] == "hello"
    assert r1.state["something"] == "hello"
    assert run_count == 1


def test_extra_metadata_used_inline(configured_test_manager):
    """Inline cacher should correctly handle extra_metadata as long as save_metadata() and
    load_metadata() are called correctly."""

    class UsesExtraMetadataCacher(Cacheable):
        def save(self, obj):
            self.extra_metadata["best_number"] = 13
            JsonCacher(self.get_path()).save(obj)
            self.save_metadata()

        def load(self):
            self.load_metadata()
            assert self.extra_metadata["best_number"] == 13
            return JsonCacher(self.get_path()).load()

    # r0 = cf.Record(configured_test_manager, cf.ExperimentArgs(name="test"))
    # UsesExtraMetadataCacher(f"{configured_test_manager.cache_path}/test_thing", record=r0).save("testing")
    UsesExtraMetadataCacher(f"{configured_test_manager.cache_path}/test_thing").save(
        "testing"
    )
    assert (
        UsesExtraMetadataCacher(
            f"{configured_test_manager.cache_path}/test_thing"
        ).load()
        == "testing"
    )



"""This module is deprecated and will be removed soon, see ``curifactory.params``."""

from dataclasses import dataclass

from curifactory.params import ExperimentParameters
from curifactory.utils import _warn_deprecation


@dataclass
class ExperimentArgs(ExperimentParameters):
    """DEPRECATED. Use ``curifactory.params.ExperimentParameters``"""

    def __post_init__(self):
        _warn_deprecation(
            "'curifactory.args.ExperimentArgs' has been deprecated and will likely be removed in 0.16.0. Please use 'curifactory.params.ExperimentParameters'"
        )

    def __init_subclass__(cls) -> None:
        _warn_deprecation(
            "'curifactory.args.ExperimentArgs' has been deprecated and will likely be removed in 0.16.0. Please use 'curifactory.params.ExperimentParameters'"
        )
        return super().__init_subclass__()



"""Contains the relevant classes for experiment state, namely the artifact manager and
record classes, as well as experiment store/parameter store management."""

import copy
import json
import logging
import multiprocessing as mp
import os
import sys
from datetime import datetime
from socket import gethostname

from curifactory import reporting, utils
from curifactory.dag import DAG
from curifactory.record import Record
from curifactory.reporting import Reportable
from curifactory.store import ManagerStore


class ArtifactManager:
    """This class manages the records, metadata, and options for an experiment run.

    Every experiment run needs an instance of this.

    Args:
        experiment_name (str): The name of the experiment. This is used as a prefix for
            most relevant information and affects caching paths. If a manager is being
            used primarily from jupyter notebooks, you could for instance set this to
            be the name of the notebook.
        store_full (bool): Store/copy environment info, log, output report, and
            all cached files in a run-specific folder (this is a --store-full run.)
        dry (bool): Setting dry to true will suppress saving any files (including logs), and will
            not update parameter stores. (It should have no effect on any files.)
        dry_cache (bool): Setting this to true only suppresses saving cache files. This is recommended
            if you're running with a cache_dir_override for some previous --store-full run, so you
            don't accidentally overwrite or add new data to the --store-full directory.
        prefix (str): Instead of using the experiment name to group cached data, use
            this name instead.
        run_line (str): The CLI command used to run the current experiment.
        parallel_lock (multiprocessing.Lock): If this function is called from a multiprocessing
            context, use this lock to help prevent files being written to and read simultaneously.
        parallel_mode (bool): This is handled by the parallel parameter, informing a particular subproc
            run that it is being executed from a parallel run.
        lazy (bool): If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have
            a cacher specified will be given a PickleCacher. Note that objects without a cacher tht do not
            handle pickle serialization correctly may cause errors.
        ignore_lazy (bool): Run the experiment disabling any lazy object caching/keeping everything in memory.
            This can save time when memory is less of an issue.
        notes (str): A git-log-like message to store in the run info for the current run. If this is an
            empty string, query the user for an input string.
        manager_cache_path (str): The path where the parameter registry and experiment
            store data are kept.
        cache_path (str): The path where all intermediate data is cached.
        runs_path (str): The path where full experiment stores (run with :code:`--store-full`)
            are kept.
        logs_path (str): The path where run logs get stored.
        notebooks_path (str): The path where run notebooks are placed.
        reports_path (str): The path where experiment run reports are saved.
        report_css_path (str): The path to a CSS file to copy into each report directory.
        status_override (str): This variable is 'LIVE' by default, but is overriden by the experiment script.
            'LIVE' indicates that this manager is being run from a non-experiment script or interactive terminal.
        suppress_live_log (bool): If true, on live/interactive runs don't automatically spawn a logger.
        live_log_debug (bool): If true, spawn live logger with the DEBUG level.
    """

    def __init__(
        self,
        experiment_name: str = None,
        store_full: bool = False,
        dry: bool = False,
        dry_cache: bool = False,
        prefix: str = None,
        run_line: str = "",
        parallel_lock: mp.Lock = None,
        parallel_mode: bool = False,
        lazy: bool = False,
        ignore_lazy: bool = False,
        notes: str = None,
        manager_cache_path: str = None,
        cache_path: str = None,
        runs_path: str = None,
        logs_path: str = None,
        notebooks_path: str = None,
        reports_path: str = None,
        report_css_path: str = None,
        status_override: str = "LIVE",
        suppress_live_log: bool = False,
        live_log_debug: bool = False,
    ):
        self.current_stage_name = ""
        """The name of the stage currently executing."""
        self.stage_active = False
        """Flag indicating whether a stage is actively running or not, used to
        detect if a stage has directly called another stage. (Technically not
        disallowed but a big no-no/prohibits any DAG features.)"""

        self.experiment_name = experiment_name
        """The name of the experiment and/or the prefix used for caching."""
        self.run_timestamp = datetime.now()
        """The datetime timestamp for when the manager is initialized (and usually
        also when the experiment starts running.)"""
        self.experiment_run_number: int = 0
        """The run counter for experiments with the given name."""
        self.parameter_files: list[str] = []
        """The list of parameter file names to be used in the experiment."""
        self.param_file_param_sets: dict[str, list[list[str, str]]] = {}
        """A dictionary of parameter file names for keys, where each value is an array of arrays,
        each inner array containing the parameter set name and the parameter set hash, for the
        parameter sets that come from that parameter file. (See the ``params`` field in the
        metadata blocks in ``ManagerStore``.)

        e.g. ``{"my_param_file": [ [ "my_param_set", "44b5e428e7165975a3e4f0d1674dbe5f" ] ] }``
        """
        self.git_commit_hash = ""
        """The current commit hash if a git repo is in use."""
        self.git_workdir_dirty = False
        """Whether there are uncommited changes in the git repo or not."""
        self.pip_freeze = ""
        """The output from a ``pip freeze`` command."""
        self.conda_env = ""
        """The output from a conda env command (``conda env export --from-history``)."""
        self.os = ""
        """The name of the current OS running curifactory."""
        self.hostname = gethostname()
        """The hostname of the machine this experiment ran on."""

        self.prefix = prefix
        """If specified, the name to use for grouping cached data instead of the experiment name."""
        self.notes = notes
        """User-entered notes associated with a session/run to output into the report etc."""

        self.manager_cache_path = manager_cache_path
        """The path where the parameter registry and experiment store data are kept."""
        self.cache_path = cache_path
        """The path where all intermediate data is cached."""
        self.runs_path = runs_path
        """The path where full experiment stores (run with ``--store-full``) are kept."""
        self.logs_path = logs_path
        """The path where run logs get stored."""
        self.notebooks_path = notebooks_path
        """The path where run notebooks are placed."""
        self.reports_path = reports_path
        """The path where experiment run reports are saved."""
        self.report_css_path = report_css_path
        """The path to a CSS file to copy into each report directory."""

        self.config = {}
        """The manager configuration loaded from the curifactory config file if present."""

        self.records = []
        """The list of records currently managed by this manager."""
        self.store_full: bool = store_full
        """Flag for whether to store/copy environment info, log, output report, and
        all cached files in a run-specific folder."""
        self.dry: bool = dry
        """Flag for whether to suppress all file outputs (cached files, logs, registries etc.)"""
        self.dry_cache: bool = dry_cache
        """Flag for whether to suppress only writing cached files."""
        self.artifacts = []
        """The list of :code:`ArtifactRepresentation` instances for all artifacts stored in all
        record states."""
        self.mapped_artifacts = []
        """The list of :code:`MapArtifactRepresentation` instances found during the mapping process."""
        self.reportables = []
        """The list of all reportables reported from all records."""

        self.stored = False
        """A flag keeping track of if the experiment :code:`store.json` has been updated with this run
        or not."""
        self.run_info = None
        """The metadata block associated with this manager from the :code:`ManagerStore`."""
        self.run_line = run_line
        """The CLI command used to run the current experiment."""
        self.reproduction_line = ""
        """The CLI command to replicate the experiment this manager is for."""
        self.interactive = False
        """Indicates if this manager was spawned with LIVE status in an interactive environment (e.g. a python
        terminal or Jupyter notebook.)"""

        # TODO: shouldn't we allow this to directly be passed into the constructor?
        self.overwrite_stages = []
        """The list of individual stages for which to ignore the cache."""
        self.overwrite = False
        """For live session managers where you don't wish to set overwrite on individual parameter sets, you
        can universely set the manager to overwrite by changing this flag to True."""

        self.error_thrown = False
        """A flag indicating whether an error was thrown by the experiment."""
        self.status = "incomplete" if status_override is None else status_override
        """The current status of the experiment: 'incomplete', 'complete', 'error', or 'LIVE' if run from a
        notebook, interactive terminal, or non-experiment script."""
        self.error = None
        """The exception class and error string, if one was thrown."""

        self.parallel_lock = parallel_lock
        """If this function is called from a multiprocessing context, use this lock to help
        prevent files being written to and read simultaneously."""
        self.parallel_mode = parallel_mode
        """A flag indicating whether this manager is in a "subproc" run or not. Do not set this manually."""

        self.lazy = lazy
        """If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have.
        a cacher specified will be given a PickleCacher. Note that objects without a cacher tht do not
        handle pickle serialization correctly may cause errors."""
        self.ignore_lazy = ignore_lazy
        """Run the experiment disabling any lazy object caching/keeping everything in memory.
        This can save time when memory is less of an issue."""

        self.live_report_path_generated = False
        """A flag indicting if the default report's graphs/reportables folders exist, this helps prevent
        live displays from breaking if multiple display functions called."""
        self.live_report_paths = None

        self.map_mode = False
        """If we're in map mode, don't actually execute any stages, we're only
        recording the 'DAG' (really just the set of stages associated with each
        record)"""
        self.map: DAG = None

        self.map_progress = None
        self.map_progress_overall_task_id = None

        self._load_config()

        if not self.dry and not self.dry_cache:
            if not os.path.exists(self.cache_path):
                os.makedirs(self.cache_path)
            if not os.path.exists(self.runs_path):
                os.makedirs(self.runs_path)

        if not self.dry:
            # TODO: (01/25/2022) maybe keep reports path check for when we're actually generating the report
            if not os.path.exists(self.reports_path):
                os.makedirs(self.reports_path)
            if not os.path.exists(self.logs_path):
                os.makedirs(self.logs_path)

        # handle populating alternate run line
        if self.status == "LIVE":
            if hasattr(sys, "ps1"):
                self.run_line = "(Interactive environment)"
                self.interactive = True
            else:
                self.run_line = " ".join(sys.argv)

        if not self.parallel_mode:
            # NOTE: by this point we don't technically have things like experiment parameters, but
            # subsequent store calls will appropriately update it (see update_run in store.py)
            # The reason this is important is so that the log init below has the correct run number etc.
            self.store()

        # start logging if from a live environment (otherwise experiment script handles this)
        if self.status == "LIVE" and not suppress_live_log and not self.parallel_mode:
            log_name = self.get_reference_name()
            log_path = os.path.join(self.logs_path, f"{log_name}.log")
            level = logging.DEBUG if live_log_debug else logging.INFO
            if self.dry:
                log_path = None
            utils.init_logging(log_path, level, False)

    # TODO: move this into dag
    def map_records(self):
        """Run through every record currently stored and grab the stage and
        stage i/o list and store it. Then clean the records."""
        self.map = DAG()

        for record in self.records:
            mapped_record = Record(self, record.params, hide=True)
            mapped_record.stages = record.stages
            mapped_record.stage_inputs = record.stage_inputs
            mapped_record.stage_outputs = record.stage_outputs
            mapped_record.stage_suppress_missing = record.stage_suppress_missing
            mapped_record.stage_kwargs_keys = record.stage_kwargs_keys
            mapped_record.stage_inputs_names = record.stage_inputs_names
            mapped_record.is_aggregate = record.is_aggregate
            mapped_record.combo_hash = record.combo_hash

            # NOTE: the state artifact representations are _not_ the actual state artifact representaitons (obviously)
            # instead we're just directly using the string names
            mapped_record.state_artifact_reps = copy.deepcopy(
                record.state_artifact_reps
            )

            self.map.records.append(mapped_record)

        # go through and add input_records now (can't directly add
        # inside previous loop because we have to add references to
        # the corresponding mapped_record instances rather than the
        # self.records instances which are about to be deleted)
        for i, record in enumerate(self.records):
            for input_record in record.input_records:
                input_record_index = self.records.index(input_record)
                self.map.records[i].input_records.append(
                    self.map.records[input_record_index]
                )

        # Move over all of the artifact representations into the map
        for artifact_rep in self.artifacts:
            self.map.artifacts.append(artifact_rep)

        self.records.clear()
        self.artifacts.clear()

        self.map.analyze()

    # update_type can either be "start" or "continue"
    # start type means "stage start", not record start, though we could check if
    # it hasn't been started yet.
    def update_map_progress(self, record: Record, update_type: str = ""):
        """Update the rich progress bar for the specified record."""
        if self.map_progress is not None and not self.map_mode:
            # self.map_progress.update(self.map_progress.task_ids[0], advance=1)
            # find appropriate record
            # TODO: use aggregate (or non) hash as the way to find the correct
            # record.
            taskid = -1
            record_name = record.get_reference_name()
            # name = record.args.name if record.args is not None else "None"
            # record_hash = record.get_hash()
            # record_index = -1
            for i, map_record in enumerate(self.map.records):
                # map_record_hash = map_record.get_hash()
                # map_name = (
                #     map_record.params.name if map_record.params is not None else "None"
                # )
                # if map_record_hash == record_hash:
                if map_record.get_reference_name(True) == record_name:
                    taskid = map_record.taskid
                    # record_index = i

            map_task = None
            for task in self.map_progress.tasks:
                if task.id == taskid:
                    map_task = task

            # continue is called when a stage/aggregate is complete and about to
            # return
            if update_type == "continue":
                self.map_progress.update(taskid, advance=1, visible=True, name="")
                if map_task.completed == map_task.total:
                    self.map_progress.update(
                        self.map_progress_overall_task_id, advance=1
                    )
                    self.map_progress.update(self.map_progress_overall_task_id, name="")

            # start gets called at the beginning of a stage or aggregate
            # decorator
            elif update_type == "start":
                self.map_progress.update(
                    taskid, visible=True, name=f"Stage {record.stages[-1]}"
                )
                if not map_task.started:
                    self.map_progress.start_task(taskid)
                self.map_progress.update(
                    self.map_progress_overall_task_id,
                    name=record_name,
                )

    def _load_config(self):
        """Populate any non-pre-existing path values with config values."""
        self.config = utils.get_configuration()
        if self.manager_cache_path is None:
            self.manager_cache_path = self.config["manager_cache_path"]
        if self.cache_path is None:
            self.cache_path = self.config["cache_path"]
        if self.runs_path is None:
            self.runs_path = self.config["runs_path"]
        if self.logs_path is None:
            self.logs_path = self.config["logs_path"]
        if self.notebooks_path is None:
            self.notebooks_path = self.config["notebooks_path"]
        if self.reports_path is None:
            self.reports_path = self.config["reports_path"]
        if self.report_css_path is None:
            self.report_css_path = self.config["report_css_path"]

    def store(self):
        """Update the ``ManagerStore`` with this manager's run metadata."""
        if self.dry:
            return
        if self.stored:
            # update
            store = ManagerStore(self.manager_cache_path)
            self.run_info = store.update_run(self)
            if self.store_full:
                # update relevant run_info too with new_run
                if self.run_info is not None:
                    with open(
                        self.get_run_output_path("run_info.json"), "w"
                    ) as outfile:
                        json.dump(self.run_info, outfile, indent=4)
            return
        else:
            store = ManagerStore(self.manager_cache_path)
            self.run_info = store.add_run(self)
            self.stored = True
            if self.store_full:
                self.write_run_env_output()

    def write_run_env_output(self):
        """Write all environment metadata to a run folder for a ``--store-full`` run."""
        self.pip_freeze = utils.get_pip_freeze()
        self.conda_env = utils.get_conda_env()
        self.os = utils.get_os()

        with open(
            self.get_run_output_path("requirements.txt"), "w", newline=""
        ) as outfile:
            outfile.write(self.pip_freeze)

        if self.conda_env != "":
            with open(
                self.get_run_output_path("environment.yml"), "w", newline=""
            ) as outfile:
                outfile.write(self.conda_env)

        with open(self.get_run_output_path("environment_meta.txt"), "w") as outfile:
            outfile.write(self.os)

        if self.run_info is not None:
            with open(self.get_run_output_path("run_info.json"), "w") as outfile:
                json.dump(self.run_info, outfile, indent=4)

    def get_all_param_sets(self) -> list:
        """This is important to get parameter sets that aren't obtained through parameter files
        , e.g. in an interactive session."""
        master_list = []
        found_hashes = []  # track which parameter sets we've already added by hash
        for record in self.records:
            if record.params is not None:
                if record.params.hash not in found_hashes:
                    master_list.append(record.params)
                    found_hashes.append(record.params.hash)

        return master_list

    def get_grouped_reportables(self) -> dict[str, list[Reportable]]:
        """Returns a dictionary of reportable groups, each group containing the list of reportables."""
        grouped_reportables = {}
        for reportable in self.reportables:
            if reportable.group is None:
                continue
            if reportable.group not in grouped_reportables:
                grouped_reportables[reportable.group] = []
            grouped_reportables[reportable.group].append(reportable)
        return grouped_reportables

    def get_ungrouped_reportables(self) -> list[Reportable]:
        """Returns the list of reportables that have no group."""
        non_grouped_reportables = []
        for reportable in self.reportables:
            if reportable.group is None:
                non_grouped_reportables.append(reportable)
        return non_grouped_reportables

    def get_artifact_path(
        self,
        obj_name: str,
        record: Record,
        subdir: str = None,
        prefix: str = None,
        stage_name: str = None,
        store: bool = False,
    ) -> str:
        """Get a record-appropriate full path/filename for a given object name and record.

        This is used by the cachers, it automatically handles generating a filename
        using appropriate experiment name prefixing etc. **NOTE:** This function sets the record's args hash if it is None, or if an aggregate stage is involved.

        The output path will follow this convention: ``[base path]/[prefix]_[parameterset hash]_[stage name]_[artifact name]``,
        where ``base path`` is determined based on the value of ``store`` and ``subdir``.

        Args:
            obj_name (str): The name to associate with the object as the last part of the filename.
            record (Record): The record that this object is associated with. (Used to get experiment name, args hash
                and so on.)
            subdir (str): An optional string of one or more nested subdirectories to prepend to the artifact filepath.
                This can be used if you want to subdivide cache and run artifacts into logical subsets, e.g. similar to
                https://towardsdatascience.com/the-importance-of-layered-thinking-in-data-engineering-a09f685edc71.
            prefix (str): An optional alternative prefix to the experiment-wide prefix (either the experiment name or
                custom-specified experiment prefix). This can be used if you want a cached object to work easier across
                multiple experiments, rather than being experiment specific. WARNING: use with caution, cross-experiment
                caching can mess with provenance.
            stage_name (str): The stage that produced an artifact. If not supplied, uses
                the currently executing stage name.
            store (bool): Set this to true if the path needs to go into a --store-full run folder.

        Returns:
            A string filepath that an object can be written to.
        """
        # TODO: provide some examples in the docstring
        args_hash = record.get_hash()
        if prefix is None:
            prefix = self._get_name()

        if stage_name is None:
            stage_name = self.current_stage_name

        # NOTE: at some point if we have better parallel handling in cf, we'll probably
        # want "current_stage_name" to be on the record level rather than the manager level.
        object_path = f"{prefix}_{args_hash}_{stage_name}_{obj_name}"

        base_path = self.cache_path
        if store:
            base_path = os.path.join(self.get_run_output_path(), "artifacts")

        # if specific subdirectories are requested, those go at the _end_ of the base path.
        # e.g. 'data/cache/my/sub/directories/[object_filepath]'
        if subdir is not None:
            base_path = os.path.join(base_path, subdir)

        # TODO: (3/21/2023) unsure if always making the path is correct, may need
        # to add a parameter for this
        os.makedirs(base_path, exist_ok=True)
        return os.path.join(base_path, object_path)

    def get_str_timestamp(self) -> str:
        """Convert the manager's run timestamp into a string representation."""
        return self.run_timestamp.strftime(utils.TIMESTAMP_FORMAT)

    def _get_name(self) -> str:
        if self.prefix is None:
            return self.experiment_name
        return self.prefix

    # NOTE: reference name is for the experiment run name, so we don't use prefix for this.
    # Prefix is only intended for caching purposes.
    def get_reference_name(self) -> str:
        """Get the reference name of this run in the experiment registry.

        The format for this name is ``[experiment_name]_[run_number]_[timestamp]``."""
        return f"{self.experiment_name}_{self.experiment_run_number}_{self.get_str_timestamp()}"

    def get_run_output_path(self, obj_name: str = None) -> str:
        """Get the path for a ``--store-full`` run folder for this manager. Similar to get_path, but
        with an always assumed output=True.

        Returns:
            A string filepath pointing to the runs folder with an optional object name attached at
            end. (Returns only the folder path if None is passed.)
        """
        output_path = os.path.join(self.runs_path, self.get_reference_name())
        return_path = output_path
        # have to be careful ordering of where we put stuff into output_path, because makedirs will
        # make _everything_ a dir, even filename a dir too, so only include if a '/' found in obj_name
        if obj_name is not None:
            return_path = os.path.join(return_path, obj_name)
            if "/" in obj_name:
                additional_sub_folders = obj_name[: obj_name.rindex("/")]
                output_path = os.path.join(output_path, additional_sub_folders)
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        return return_path

    def lock(self):
        """Necessary to avoid file collisions when run in parallel mode."""
        if self.parallel_lock is not None:
            self.parallel_lock.acquire()

    def unlock(self):
        """Necessary to avoid file collisions when run in parallel mode."""
        if self.parallel_lock is not None:
            self.parallel_lock.release()

    def generate_report(self):
        """Output report files to a run-specific report folder, the reports/_latest, and the
        store-full folder if applicable."""
        if self.dry:
            return
        logging.info("Generating report...")

        self.store()

        reporting.run_report(self, self.reports_path, "_latest", self.report_css_path)
        self.live_report_paths = reporting.run_report(
            self, self.reports_path, self.get_reference_name(), self.report_css_path
        )
        self.live_report_path_generated = True
        if self.store_full:
            reporting.run_report(
                self, self.get_run_output_path(), "report", self.report_css_path
            )
        reporting.update_report_index(
            self.config["experiments_module_name"], self.reports_path
        )

    def display_info(self):
        """Returns an IPython HTML display object with the report info block. This is mostly
        only useful for making displays in a Juptyer notebook."""
        try:
            from IPython.display import HTML
        except ModuleNotFoundError:
            return "Unable to import IPython."
        self.store()

        return HTML("".join(reporting.render_report_info_block(self)))

    def _reportable_display_prep(self):
        try:
            from IPython.display import HTML
        except ModuleNotFoundError:
            return "Unable to import IPython."

        self.store()

        if not self.live_report_path_generated:
            self.live_report_paths = reporting.prepare_report_path(
                self.reports_path, self.get_reference_name()
            )
            self.live_report_path_generated = True
        return HTML

    def display_all_reportables(self):
        """Returns an HTML IPython display all produced reportables.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html = reporting.render_report_all_reportables(
            self,
            reportables_path,
            override_display_path=os.path.join("/", "files", reportables_path),
            notebook=True,
        )

        return HTML("".join(html))

    def display_record_reportables(self, record: Record):
        """Returns an HTML IPython display of all reportables produced
        within the passed record.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.record == record:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_group_reportables(self, group_name: str):
        """Returns an HTML IPython display of all reportables in the passed group.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.group == group_name:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_stage_reportables(self, stage_name: str):
        """Returns an HTML IPython display of all reportables produced by the given stage.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html_lines = []
        for reportable in self.reportables:
            if reportable.stage == stage_name:
                html_lines.extend(
                    reporting.render_reportable(
                        reportable,
                        self,
                        reportables_path,
                        override_display_path=os.path.join(
                            "/", "files", reportables_path
                        ),
                        notebook=True,
                    )
                )

        return HTML("".join(html_lines))

    def display_reportable(self, reportable: Reportable):
        """Returns an HTML IPython display of the rendered passed reportable.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()
        folder_path, graphs_path, reportables_path = self.live_report_paths

        html = reporting.render_reportable(
            reportable,
            self,
            reportables_path,
            override_display_path=os.path.join("/", "files", reportables_path),
            notebook=True,
        )

        return HTML("".join(html))

    def display_stage_graph(self):
        """Returns an HTML IPython display of the graphviz SVG map of the records and the stages they were run through.

        Note:
            This only works within an IPython context.
        """
        HTML = self._reportable_display_prep()

        html = reporting.render_graph(reporting.map_full_svg(self))
        return HTML("".join(html))

    def get_reportable_groups(self) -> list[str]:
        groups = []
        for reportable in self.reportables:
            if reportable.group not in groups:
                groups.append(reportable.group)
        return groups



"""Testing the `curifactory init` command."""

import json
import os

import pytest
from pytest_mock import mocker  # noqa: F401 -- flake8 doesn't see it's used as fixture

from curifactory.project import initialize_project
from curifactory.utils import get_configuration


@pytest.mark.noautofixt
def test_project_init_defaults(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """The default folders should be created when no additional input given."""

    def actual_readline():
        return "\n"

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    assert os.path.exists("data")
    assert os.path.exists("docker")
    assert os.path.exists("experiments")
    assert os.path.exists("logs")
    assert os.path.exists("notebooks")
    assert os.path.exists("notebooks/experiments")
    assert os.path.exists("params")
    assert os.path.exists("reports")
    assert os.path.exists("reports/style.css")
    assert os.path.exists(".gitignore")
    assert os.path.exists("curifactory_config.json")


@pytest.mark.noautofixt
def test_project_init_input_nondefaults(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """Passing a different value to one of the inputs should change the
    directory and config."""

    prompt_number = 0

    def actual_readline():
        nonlocal prompt_number
        prompt_number += 1
        if prompt_number == 1:
            return "myexperiments\n"
        elif prompt_number == 2:
            return "myparams\n"
        elif prompt_number == 3:
            return "mydata\n"
        elif prompt_number == 4:
            return "mylogs\n"
        elif prompt_number == 5:
            return "mynotebooks\n"
        elif prompt_number == 6:
            return "myreports\n"
        elif prompt_number == 7:  # docker check
            return "n\n"
        elif prompt_number == 8:  # .git init check
            return "\n"  # don't handle for now, difficult because this is running in a folder that already is in git repo
        elif prompt_number == 9:  # gitignore check
            return "n\n"
        else:
            return "\n"

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    assert not os.path.exists("experiments")
    assert os.path.exists("myexperiments")
    assert not os.path.exists("data")
    assert os.path.exists("mydata")
    assert not os.path.exists("docker")
    assert not os.path.exists("logs")
    assert os.path.exists("mylogs")
    assert not os.path.exists("notebooks")
    assert os.path.exists("mynotebooks")
    assert os.path.exists("mynotebooks/experiments")
    assert not os.path.exists("params")
    assert os.path.exists("myparams")
    assert not os.path.exists("reports")
    assert os.path.exists("myreports")
    assert os.path.exists("myreports/style.css")
    assert not os.path.exists(".gitignore")
    assert os.path.exists("curifactory_config.json")

    # TODO: (03/07/2022) assert that config is correct


@pytest.mark.noautofixt
def test_project_init_preexisting(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """A pre-existing curifactory config should correctly propagate values
    during init."""

    def actual_readline():
        return "\n"

    config = get_configuration()
    config["experiments_module_name"] = "myexperiments"
    config["params_module_name"] = "myparams"
    config["manager_cache_path"] = "mydata/"
    config["cache_path"] = "mydata/cache"
    config["runs_path"] = "mydata/runs"
    config["logs_path"] = "mylogs"
    config["notebooks_path"] = "mynotebooks/"
    config["reports_path"] = "myreports/"
    config["reports_css_path"] = "myreports/style.css"

    with open("curifactory_config.json", "w") as outfile:
        json.dump(config, outfile)

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    assert not os.path.exists("experiments")
    assert os.path.exists("myexperiments")
    assert not os.path.exists("data")
    assert os.path.exists("mydata")
    assert not os.path.exists("logs")
    assert os.path.exists("mylogs")
    assert not os.path.exists("notebooks")
    assert os.path.exists("mynotebooks")
    assert os.path.exists("mynotebooks/experiments")
    assert not os.path.exists("params")
    assert os.path.exists("myparams")
    assert not os.path.exists("reports")
    assert os.path.exists("myreports")
    assert os.path.exists("curifactory_config.json")


@pytest.mark.noautofixt
def test_empty_gitignore_no_blank_line(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """A blank or nonexisting gitignore should not have a blank line added
    before the curifactory section."""

    def actual_readline():
        return "\n"

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    with open(".gitignore") as infile:
        lines = infile.readlines()

    assert lines[0] == "# curifactory paths\n"


@pytest.mark.noautofixt
def test_nonempty_gitignore_has_blank_line(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """If there's a pre-existing gitignore, the .gitignore append on init should
    add a blank line before the curifactory paths, for optimal neatness."""

    def actual_readline():
        return "\n"

    with open(".gitignore", "w") as outfile:
        outfile.writelines(["stuff\n"])

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    with open(".gitignore") as infile:
        lines = infile.readlines()

    assert lines[0] == "stuff\n"
    assert lines[1] == "\n"
    assert lines[2] == "# curifactory paths\n"


@pytest.mark.noautofixt
def test_nonempty_gitignore_has_only_one_blank_line(
    mocker, project_folder  # noqa: F811 -- mocker has to be passed in as fixture
):
    """If there's a pre-existing gitignore, the .gitignore append on init should
    add a blank line before the curifactory paths, for optimal neatness, unless
    there's already a blank line there."""

    def actual_readline():
        return "\n"

    with open(".gitignore", "w") as outfile:
        outfile.writelines(["stuff\n\n"])

    mocker.patch("sys.stdin.readline", actual_readline)
    initialize_project()

    with open(".gitignore") as infile:
        lines = infile.readlines()

    assert lines[0] == "stuff\n"
    assert lines[1] == "\n"
    assert lines[2] == "# curifactory paths\n"



"""This is the 'main' runnable function and CLI, which handles setting up logging,
folders, reports, and running the passed experiment.

This file contains a :code:`__name__ == "__main__"` and can be run directly.
"""

import argparse
import datetime
import importlib
import logging
import multiprocessing as mp
import os
import re
import subprocess
import sys

import argcomplete

from curifactory import reporting, utils
from curifactory.manager import ArtifactManager

CONFIGURATION_FILE = "curifactory_config.json"


def run_experiment(  # noqa: C901 -- TODO: this does need to be broken up at some point
    experiment_name,
    param_files,
    overwrite_override=None,
    cache_dir_override=None,
    mngr: ArtifactManager = None,
    log: bool = False,
    log_debug: bool = False,
    dry: bool = False,
    dry_cache: bool = False,
    store_full: bool = False,
    log_errors: bool = False,
    prefix: str = None,
    build_docker: bool = False,
    build_notebook: bool = False,
    run_string: str = None,
    stage_overwrites: list[str] = None,
    param_set_names: list[str] = None,
    param_set_indices: list[str] = None,
    global_param_set_indices: list[str] = None,
    parallel: int = None,
    parallel_mode: bool = False,
    parallel_lock: mp.Lock = None,
    parallel_queue: mp.Queue = None,
    run_num_override: int = None,
    run_ts_override: datetime.datetime = None,
    lazy: bool = False,
    ignore_lazy: bool = False,
    no_dag: bool = False,
    map_only: bool = False,
    no_color: bool = False,
    quiet: bool = False,
    progress: bool = False,
    plain: bool = False,
    notes: str = None,
):
    """The experiment entrypoint function. This executes the given experiment
    with the given parameters.

    Args:
        experiment_name (str): The name of the experiment script (without the ``.py``).
        param_files (List[str]): A list of names of parameter files (without the ``.py``).
        overwrite_override (bool): Whether to force overwrite on all cache data.
        cache_dir_override (str): Specify a non-default cache location. This would be used if
            running with the cache from a previous ``--store-full`` run.
        mngr (ArtifactManager): An artifact manager to use for the experiment. One will be automatically
            created if none is passed.
        log (bool): Whether to write a log file or not.
        log_debug (bool): Whether to include DEBUG level messages in the log.
        dry (bool): Setting dry to true will suppress saving any files (including logs), and will
            not update parameter stores. (It should have no effect on any files.)
        dry_cache (bool): Setting this to true only suppresses saving cache files. This is recommended
            if you're running with a cache_dir_override for some previous ``--store-full`` run, so you
            don't accidentally overwrite or add new data to the ``--store-full`` directory.
        store_full (bool): Store environment info, log, output report, and all cached files in a
            run-specific folder (:code:`data/runs` by default)
        log_errors (bool): Whether to include error messages in the log output.
        prefix (str): Instead of using the experiment name to group cached data, use this prefix instead.
        build_docker (bool): If true, build a docker image with all of the run cache afterwards.
        build_notebook (bool): If true, add a notebook with run info and default cells to reproduce
            after run execution.
        run_string (str): An automatically populated string representing the CLI command for the run,
            do not change this.
        stage_overwrites (List[str]): A list of string stage names that you should overwrite, this is
            useful if there are specific stages you sometimes want to recompute but the remainder of
            the data can remain cached.
        param_set_names (List[str]): A list of parameter set names to run. If this is specified, only parameter
            sets with these names will be passed on to the experiment.
        param_set_indices (List[str]): A list of parameter set indices to run. If this is specified, only the
            parameter sets returned from the passed parameter files, indexed by the ranges specified, will
            be passed to the experiment. Note that you can specify ranges delineated with '-', e.g. '3-7'.
        global_param_set_indices (List[str]): A list of parameter set indices to run, indexing the entire collection
            of parameter sets passed to the experiment instead of each individual parameters file. This
            can be used to help more intelligently parallelize runs. Formatting follows the same rules
            as param_set_indices.
        parallel (int): How many subprocesses to split this run into. If specified, the experiment will
            be run that many times with divided up global_param_set_indices in order to generated cached
            data for all parameters, and then re-run a final time with all cached data combined. Note
            then that any speedup from this is based on how well and how many steps are cached.
        parallel_mode (bool): This is handled by the parallel parameter, informing a particular subproc
            run that it is being executed from a parallel run.
        parallel_lock (multiprocessing.Lock): If this function is called from a multiprocessing context, use
            this lock to help prevent files being written to and read simultaneously.
        parallel_queue (multiprocessing.Queue): If this function is called from a multiprocessing context,
            use this queue for communicating success/errors back to the main process.
        run_num_override (int): Handled in parallel mode, since parallel process experiments do not get
            run number (since manager.store is not called) the log gets stored in an incorrectly named
            file.
        run_ts_override (datetime.datetime): Handled in parallel mode, since parallel process experiments
            do not get the same timestamp if started a few seconds later, the log gets stored in an
            incorrectly named file.
        lazy (bool): If true, attempts to set all stage outputs as Lazy objects. Outputs that do not have
            a cacher specified will be given a ``PickleCacher``. Note that objects without a cacher that do not
            handle pickle serialization correctly may cause errors.
        ignore_lazy (bool): Run the experiment disabling any lazy object caching/keeping everything in memory.
            This can save time when memory is less of an issue.
        no_dag (bool): Prevent pre-execution mapping of experiment records and stages. Recommended if doing
            anything fancy with records like dynamically creating them based on results of previous records.
            Mapping is done by running the experiment but skipping all stage execution.
        map_only (bool): Runs the pre-execution mapping of an experiment and immediately exits, printing the
            map to stdout. **Note that setting this to True automatically sets dry.**
        no_color (bool): Suppress fancy colors in console output.
        quiet (bool): Suppress all console log output.
        progress (bool): Display fancy rich progress bars for each record.
        plain (bool): Use normal text log output rather than rich log. Note that
            this negates progress.
        notes (str): A git-log-like message to store in the run info for the current run. If this is an
            empty string, query the user for an input string.

    Returns:
        Whatever is returned from the experiment :code:`run()`.

    Example:
        .. code-block:: python

            from curifactory import experiment, ArtifactManager
            mngr = ArtifactManager()
            experiment.run_experiment('exp_name', ['params1', 'params2'], mngr=mngr, dry=True)
    """

    # doing imports here to avoid needing such a long import list for tab completion functionality
    import glob
    import shutil
    import traceback

    from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn

    from curifactory import docker, hashing

    # if we request a map only run, make sure we don't impact any files, so
    # automatically set "dry"
    if map_only:
        dry = True

    if run_string is None:
        run_string = f"experiment {experiment_name}"
        for param_file in param_files:
            run_string += f" -p {param_file}"
        if param_set_names is not None:
            for param_set_name in param_set_names:
                run_string += f" --names {param_set_name}"
        if param_set_indices is not None:
            for param_set_index in param_set_indices:
                run_string += f" --indices {param_set_index}"
        if stage_overwrites is not None:
            for stage_overwrite in stage_overwrites:
                run_string += f" --overwrite-stage {stage_overwrite}"
        if cache_dir_override is not None:
            run_string += f" --cache {cache_dir_override}"
        if overwrite_override:
            run_string += " --overwrite"
        if parallel is not None:
            run_string += f" --parallel {parallel}"
        if parallel_mode:
            run_string += " --parallel-safe"
        if global_param_set_indices is not None:
            for index_range in global_param_set_indices:
                run_string += f" --global-indices {index_range}"
        if store_full:
            run_string += " --store-full"
        if dry:
            run_string += " --dry"
        if dry_cache:
            run_string += " --dry-cache"
        if prefix is not None:
            run_string += f" --prefix {prefix}"
        if lazy:
            run_string += " --lazy"
        if ignore_lazy:
            run_string += " --ignore-lazy"
        if no_dag:
            run_string += " --no-dag"
        if map_only:
            run_string += " --map"
        if no_color:
            run_string += " --no-color"
        if quiet:
            run_string += " --quiet"

        # TODO: remainder of flags

    # NOTE: this allows a project root /params /experiments to work, but unfortunately doesn't automagically work if you have nested folders (you need to put __init__.py in each dir)
    sys.path.append(os.getcwd())

    # distributed run check, automatically set parallel mode if we're not rank 0
    # this was added because of issues with pytorch distributed compute
    distributed_mode_detected = False
    warn_store_full_during_distributed = False
    if "LOCAL_RANK" in os.environ and os.getenv("LOCAL_RANK") != "0":
        distributed_mode_detected = True
    elif "NODE_RANK" in os.environ and os.getenv("NODE_RANK") != "0":
        distributed_mode_detected = True
    elif "RANK" in os.environ and os.getenv("RANK") != "0":
        distributed_mode_detected = True
    if distributed_mode_detected:
        parallel_mode = True

        # if we're in a distributed run, we need to make sure to _not_ do
        # a store full on the non-rank-0 processes, otherwise each process
        # outputs a full copy into a full store folder that isn't even labeled
        # correctly (doesn't have run info with run num.) We throw a warning
        # to ensure the user knows any full store caching they do (really all
        # caching) needs to be handled by the rank 0 process.
        if store_full:
            warn_store_full_during_distributed = True
            store_full = False

    # get the experiment run notes if requested
    if notes == "":
        notes_file = "temp_run_notes_content.txt"
        # create notes template
        with open(notes_file, "w") as outfile:
            outfile.write(
                "\n\n"
                + '# Write a "git-commit-like" message to attach as notes to the experiment run. The first line will be treated as the shortform notes line.'
                + "\n# All lines starting with a '#' will be ignored."
            )

        # try to find an appropriate editor
        os.system(f"{utils.get_editor()} {notes_file}")

        # get the content
        with open(notes_file) as infile:
            lines = [line for line in infile.readlines() if not line.startswith("#")]
        notes = "".join(lines).strip("\n")

        os.remove(notes_file)

    # force full store if building a docker container
    if build_docker:
        store_full = True

    if mngr is None:
        mngr = ArtifactManager(
            experiment_name,
            store_full=store_full,
            dry=dry,
            dry_cache=dry_cache,
            prefix=prefix,
            run_line=run_string,
            parallel_lock=parallel_lock,
            parallel_mode=parallel_mode,
            lazy=lazy,
            ignore_lazy=ignore_lazy,
            status_override=None,
            notes=notes,
        )
        # mngr.experiment_name = experiment_name
        mngr.parameter_files = param_files
        if run_num_override is not None:
            mngr.experiment_run_number = run_num_override
        if run_ts_override is not None:
            mngr.run_timestamp = run_ts_override

        if stage_overwrites is not None:
            # TODO: only apply to subprocs if parallel?
            mngr.overwrite_stages = stage_overwrites

    # check for cache path override
    if cache_dir_override is not None:
        mngr.cache_path = cache_dir_override

    # resolve args indices and any specified ranges
    param_set_indices_resolved = []
    if param_set_indices is not None:
        for entry in param_set_indices:
            if "-" in entry:
                start = int(entry[: entry.index("-")])
                end = int(entry[entry.index("-") + 1 :])
                param_set_indices_resolved.extend(list(range(start, end)))
            else:
                param_set_indices_resolved.append(int(entry))

    # resolve global args indices and any specified ranges
    global_param_set_indices_resolved = []
    if global_param_set_indices is not None:
        for entry in global_param_set_indices:
            if "-" in entry:
                start = int(entry[: entry.index("-")])
                end = int(entry[entry.index("-") + 1 :])
                global_param_set_indices_resolved.extend(list(range(start, end)))
            else:
                global_param_set_indices_resolved.append(int(entry))

    # handling logging. NOTE: we do have to keep this here because of parallel considerations
    if log:
        log_name = mngr.get_reference_name()
        if parallel_mode:
            if global_param_set_indices is not None:
                log_name += f"_{str(global_param_set_indices[0])}"
            else:
                log_name += f"_p{str(os.getpid())}"
        log_path = os.path.join(mngr.logs_path, f"{log_name}.log")

        if dry:
            log_path = None

        level = logging.INFO
        if log_debug:
            level = logging.DEBUG
        utils.init_logging(
            log_path,
            level,
            log_errors,
            include_process=parallel_mode,
            quiet=quiet,
            no_color=no_color,
            plain=plain,
        )

    logging.info("Running experiment %s" % experiment_name)
    if mngr.git_workdir_dirty:
        logging.warning(
            "Git working directory contains uncommited changes. Reproduction may not be exact."
        )

    # insert an implied parameter file that is the same as the experiment file, if no parameters were explicitly listed
    if param_files is None or len(param_files) == 0:
        logging.info(
            "No parameter files listed, assuming an implied experiment file get_params(), inserting '%s'"
            % experiment_name
        )
        param_files = [experiment_name]

    # let the user that using "map-only" implies dry
    if map_only:
        logging.info(
            "Using 'map-only' mode - note that this implies --dry and no cache or store files will be modified."
        )

    # let the user know if we detected a distributed run
    if distributed_mode_detected:
        logging.info(
            "A distributed run was detected from environment variables - automatically using --parallel-mode on non-rank-0 processes."
        )

    # if the user was doing a full store distributed run, let them know the non-rank-zero processes
    # won't be storing anything in the full store directory
    if warn_store_full_during_distributed:
        logging.warning(
            "Full store was requested for a distributed run - full store has been disabled on all non-rank-zero processes to prevent data duplication. Ensure that any necessary data for a full store is handled in rank zero."
        )

    # load params files
    final_param_sets = []
    for param_file_name in param_files:
        mngr.param_file_param_sets[param_file_name] = []
        # TODO: if we get a ModuleNotFoundError, suggest ensuring __init__.py as appropriate and that module paths don't have '/'
        param_module_string = f"{mngr.config['params_module_name']}.{param_file_name}"
        experiment_module_string = (
            f"{mngr.config['experiments_module_name']}.{param_file_name}"
        )
        try:
            logging.debug("Trying to load params module '%s'" % param_module_string)
            param_module = importlib.import_module(param_module_string)
        except ModuleNotFoundError:
            try:
                logging.debug(
                    "Module not found, trying '%s'" % experiment_module_string
                )
                param_module = importlib.import_module(experiment_module_string)
            except ModuleNotFoundError as e:
                logging.error(
                    "Parameter file '%s' could not be found in either experiments or parameters directory. Ensure curifactory_config.json is correct and if module subpaths are used, try including an __init__.py in each folder."
                    % param_file_name
                )
                raise e

        param_sets = param_module.get_params()
        if type(param_sets) != list:
            logging.error(
                "Parameter file '%s' did not return a list, please make sure any `get_params()` functions are returning non-empty arrays."
                % param_file_name
            )
            raise RuntimeError(
                "Parameter file '%s' did not return a list, please make sure any `get_params()` functions are returning non-empty arrays."
                % param_file_name
            )
        param_sets_to_add = []

        # NOTE: we don't want to set override in parent proc on parallel runs.
        if overwrite_override and parallel is None:
            for argset in param_sets:
                argset.overwrite = True

        # compute the hash of every argset and store the params
        for index, param_set in enumerate(param_sets):
            # if specific names requested, just grab those
            if param_set_names is not None:
                if param_set.name not in param_set_names:
                    continue

            # if specific indices requested, just grab those
            if len(param_set_indices_resolved) > 0:
                if index not in param_set_indices_resolved:
                    continue

            param_set.hash = hashing.hash_param_set(
                param_set,
                store_in_registry=(not dry and not parallel_mode),
                registry_path=mngr.manager_cache_path,
            )
            mngr.param_file_param_sets[param_file_name].append(
                (param_set.name, param_set.hash)
            )
            # TODO: (01/24/2022) I have no idea what the point of this args_hash is...
            # it's not storing anything, and args_hash has no side-effects, so unclear
            # on why this matters.
            # NOTE: (3/5/2023) is it just so that the hash is computed and stored
            # on the args? Unclear exactly why that's necessary but that is technically
            # a side-effect
            if store_full:
                hashing.hash_param_set(
                    param_set, store_in_registry=False
                )  # don't try to store because get_run_output_path does not exist yet
            param_sets_to_add.append(param_set)

        final_param_sets.extend(param_sets_to_add)

    # check that there wasn't an invalid name and all requested parameterset names were found
    if param_set_names is not None:
        for param_set_name in param_set_names:
            found = False
            for param_set in final_param_sets:
                if param_set_name == param_set.name:
                    found = True
                    break
            if not found:
                logging.error(
                    "Paramset name '%s' not found in any of the provided parameter files."
                    % param_set_name
                )
                raise RuntimeError(
                    "Paramset name '%s' not found in any of the provided parameter files."
                    % param_set_name
                )

    # check that we actually have parameters
    if len(final_param_sets) == 0:
        logging.error(
            "No parameter sets found, please make sure any `get_params()` functions are returning non-empty arrays."
        )
        raise RuntimeError(
            "No parameter sets found, please make sure any `get_params()` functions are returning non-empty arrays."
        )

    if len(global_param_set_indices_resolved) > 0:
        final_param_sets = [
            final_param_sets[i] for i in global_param_set_indices_resolved
        ]

    if not parallel_mode:
        mngr.store()

    if not parallel_mode:
        # this has to be done AFTER mngr.store, otherwise the reference name does not have the correct run number.
        logging.info(
            "Experiment run reference name is '%s'" % mngr.get_reference_name()
        )
    logging.info("Run command is '%s'" % run_string)

    # store params registry in run folder
    # TODO: is this already being done in manager get_path?
    if store_full:
        for param_set in final_param_sets:
            hashing.hash_param_set(
                param_set,
                store_in_registry=(not dry and not parallel_mode),
                registry_path=mngr.get_run_output_path(),
            )

    # note that nothing is being cached or stored
    if dry:
        logging.info(
            "NOTE - running in dry mode. This log will not be saved, and no new"
            + " files will be cached. (Existing caches will still be read.)"
        )
    if store_full:
        logging.info(
            "NOTE - running in full store mode. A copy of all cached objects and environment information will be stored in %s",
            mngr.get_run_output_path(),
        )

    # handle parallelism
    if parallel is not None:
        logging.info("Running experiment in parallel on %s processes..." % parallel)
        # divide up args list
        countper = int(len(final_param_sets) / parallel)
        counts = [countper] * parallel
        remainder = len(final_param_sets) % parallel
        index = 0
        while remainder > 0:
            counts[index] += 1
            remainder -= 1
            index += 1
            if index >= len(final_param_sets):
                index = 0

        # get the formatted string range for each thread
        global_index_ranges = []
        start = 0
        for count in counts:
            end = start + count
            if count == 0:
                logging.warning(
                    "More processes specified than argsets, will run with fewer processes."
                )
            else:
                global_index_ranges.append([f"{start}-{end}"])
            start = end
        logging.info("Parallel arg splits: %s", str(global_index_ranges))

        processes = []
        multiprocessing_lock = mp.Lock()
        multiprocessing_queue = mp.Queue()
        for index_range in global_index_ranges:
            p = mp.Process(
                target=run_experiment,
                args=(
                    experiment_name,
                    param_files,
                    overwrite_override,
                    cache_dir_override,
                    None,  # mngr
                    log,
                    log_debug,
                    dry,
                    dry_cache,
                    False,  # store_full
                    log_errors,
                    prefix,
                    False,  # build_docker
                    False,  # build_notebook
                    None,  # run_string
                    stage_overwrites,
                    param_set_names,  # param_set_names
                    param_set_indices,  # param_set_indices
                    index_range,
                    None,  # parallel
                    True,  # parallel_mode
                    multiprocessing_lock,
                    multiprocessing_queue,
                    mngr.experiment_run_number,
                    mngr.run_timestamp,
                    lazy,
                    ignore_lazy,
                    no_dag,
                    map_only,
                    no_color,
                    quiet,
                    progress,
                    plain,
                ),
            )
            logging.info(
                "Starting new experiment process for parameter range %s...", index_range
            )
            p.start()
            processes.append(p)

        for i, p in enumerate(processes):
            # TODO - get success/fail status
            p.join()
            run_status = multiprocessing_queue.get()
            if run_status[1] == "error":
                mngr.status = "error"
                mngr.error = run_status[2]
                mngr.error_thrown = True
                logging.info(
                    "Joined process for range '%s' with status code '%s' - %s"
                    % (global_index_ranges[i], run_status[1], run_status[2])
                )
            else:
                logging.info(
                    "Joined process for range '%s' with status code '%s'"
                    % (global_index_ranges[i], run_status[1])
                )

        logging.info(
            "All experiment processes have re-joined, continuing final combined run..."
        )

    if parallel_mode:
        logging.info(
            "Running experiment process for parameter range %s"
            % str(global_param_set_indices)
        )

    # run the experiment
    error_thrown = False
    experiment_module = importlib.import_module(
        f"{mngr.config['experiments_module_name']}.{experiment_name}"
    )
    try:
        # run experiment mapping and set up progress bars
        if not parallel_mode and not no_dag:
            logging.info("Pre-mapping stages and records")
            mngr.map_mode = True
            experiment_module.run(final_param_sets, mngr)
            mngr.map_mode = False
            logging.debug("Constructing record map")
            mngr.map_records()
            for tree in mngr.map.execution_trees:
                logging.debug(str(tree))
            logging.debug("Execution list: %s" % mngr.map.execution_list)
            # logging.debug("Execution map: %s" % mngr.map.execution_chain)
            logging.info("Stage map collected")

            if map_only:
                dag = mngr.map
                dag.print_experiment_map()
                logging.info("Map-only mode, skipping remainder of experiment.")
                return dag, mngr

            # create a (rich) progress bar and the associated tasks for each
            # mapped record.
            if progress:
                mngr.map_progress = Progress(
                    TextColumn("{task.completed}/{task.total}"),
                    BarColumn(bar_width=30, pulse_style="cyan"),
                    TextColumn("[progress.description]{task.description}"),
                    TimeElapsedColumn(),
                    TextColumn("{task.fields[name]}"),
                )
                for i, record in enumerate(mngr.map.records):
                    name = record.params.name if record.params is not None else "None"
                    if record.is_aggregate:
                        name += " (aggregate)"
                    hash_big = record.get_hash()
                    hash_cut = record.get_hash()[:8]
                    # logging.debug("LENGTH %s" % len(record.stages))
                    taskid = mngr.map_progress.add_task(
                        f"Record {i}, args: {name}, hash: {hash_cut}",
                        total=len(record.stages),
                        start=False,
                        visible=True,
                        name="",
                        hash_big=hash_big,
                        hash_cut=hash_cut,
                    )
                    record.taskid = taskid
                overalltaskid = mngr.map_progress.add_task(
                    f"Total ({mngr.get_reference_name()})",
                    total=len(mngr.map.records),
                    visible=True,
                    name="",
                    hash_big="",
                    hash_cut="",
                )
                mngr.map_progress_overall_task_id = overalltaskid

                mngr.map_progress.start()

        results = experiment_module.run(final_param_sets, mngr)

        if not parallel_mode and not no_dag and progress:
            mngr.map_progress.stop()

        # don't change status if we logged an error from a parallel process
        if not mngr.error_thrown:
            mngr.status = "complete"
    except Exception as e:
        if not parallel_mode and not no_dag and progress:
            mngr.map_progress.stop()
        results = None
        error_thrown = True
        logging.error(e)
        logging.error(traceback.format_exc())

        # don't change status if we logged an error from a parallel process
        if not mngr.error_thrown:
            mngr.status = "error"
            mngr.error = f"{str(e.__class__.__name__)} - {str(e)}"

    # send the results from a parallel process to the main process
    if parallel_mode and parallel_queue is not None:
        parallel_queue.put((global_param_set_indices[0], mngr.status, mngr.error))

    if not parallel_mode:
        mngr.store()

    if log and store_full and not dry:
        # copy the logfile(s) over (when running in parallel, grab the subproc logs too)
        log_name = mngr.get_reference_name()
        log_path = f"logs/{log_name}*.log"
        for file in glob.glob(log_path):
            shutil.copy(file, mngr.get_run_output_path())
        # shutil.copyfile(
        #     log_path, os.path.join(mngr.get_run_output_path(), f"{log_name}.log")
        # )

        # TODO details file for any meta information (if it errored or not)

    if build_notebook and not dry and not parallel_mode:
        notebook_loc = os.path.join(mngr.notebooks_path, "experiments")
        try:
            os.makedirs(notebook_loc, exist_ok=True)
        # TODO: Handle exceptions better
        except:  # noqa: E722
            pass
        write_experiment_notebook(
            experiment_name,
            param_files,
            final_param_sets,
            mngr,
            path=os.path.join(notebook_loc, mngr.get_reference_name()),
            use_global_cache=None,
            errored=error_thrown,
        )

    if build_docker and not dry and not parallel_mode:
        if error_thrown:
            logging.warning("The experiment threw an error - docker build skipped.")
        else:
            write_experiment_notebook(
                experiment_name,
                param_files,
                final_param_sets,
                mngr,
                mngr.get_run_output_path("run_notebook"),
                use_global_cache=True,
                directory_change_back_depth=1,
                suppress_global_warning=True,
            )
            docker.build_docker(
                experiment_name,
                mngr.get_run_output_path(),
                f"{mngr.experiment_run_number}_{mngr.run_timestamp.strftime('%Y-%m-%d')}",
            )

    if not dry and not parallel_mode:
        mngr.generate_report()

    return results, mngr


def write_experiment_notebook(
    experiment_name,
    param_files,
    argsets,
    manager,
    path,
    directory_change_back_depth=2,
    use_global_cache=None,
    errored=False,
    suppress_global_warning=False,
):
    """Creates a jupyter notebook prepopulated with experiment info and cells to re-run
    the experiment and discover all associated data stored in record states. This function
    is run by the :code:`run_experiment()` function.

    Args:
        experiment_name (str): The name of the run experiment
        param_files (List[str]): List of parameter file names
        argsets (List[Args]): List of all used :code:`Args` from parameter files.
        manager (ArtifactManager): :code:`ArtifactManager` used in the experiment.
        path (str): The path to the directory to store the notebook in.
        directory_change_back_depth (int): How many directories up the notebook needs
            to be in the project root (so imports and cache paths are all correct.)
        use_global_cache (bool): Whether we're using the normal experiment cache or
            a separate specific cache folder (mostly just used to display a warning
            in the notebook.)
        errored (bool): Whether the experiment errored or nat while running, will display
            a warning in the notebook.
        suppress_global_warning (bool): Don't show a warning if :code:`use_global_cache`
            is true.
    """
    # TODO - move elsewhere?
    logging.info("Creating experiment notebook...")

    if use_global_cache is None:
        use_global_cache = not manager.store_full

    output_lines = [
        "# %%",
        "'''",
        f"# {manager.experiment_name} - {manager.experiment_run_number}",
        f"\nExperiment name: **{manager.experiment_name}**  ",
        f"Experiment run number: **{manager.experiment_run_number}**  ",
        f"Run timestamp: **{manager.run_timestamp.strftime('%m/%d/%Y %H:%M:%S')}**  ",
        f"Reference: **{manager.get_reference_name()}**  ",
        f"Git commit: {manager.git_commit_hash}  ",
        f"Param files: {str(manager.parameter_files)}",
        "\n**Parameters**:\n",
    ]

    # output the list of parameters used and assoc hashes
    for key in manager.param_file_param_sets:
        output_lines.append(f"* {key}")
        for name, hash in manager.param_file_param_sets[key]:
            output_lines.append(f"\t* {name} - {hash}")

    output_lines.extend(
        [
            "---",
            "'''",
            "",
            # "# %%",
            # "'''",
        ]
    )

    # NOTE: for large argsets it makes this cell take up way too much room
    # # output detailed args for each paramset
    # for key in manager.experiment_args:
    #     for name, hash in manager.experiment_args[key]:
    #         output_lines.append(f"\n**{name} - {hash}**")
    #         output_lines.append("```")
    #         for argset in argsets:
    #             if argset.hash == hash:
    #                 argset_data = asdict(argset)
    #                 del argset_data["name"]
    #                 del argset_data["overwrite"]
    #                 del argset_data["hash"]
    #                 output_lines.append(
    #                     json.dumps(argset_data, indent=4, default=lambda x: str(x))
    #                 )
    #                 break
    #         output_lines.append("```")

    # output_lines.extend(["---", "'''", ""])

    # pathing for whether full run or not
    directory_change_back = "/".join([".."] * directory_change_back_depth)
    cache_dir_arg = f"manager_cache_path='{manager.get_run_output_path()}', cache_path='{manager.get_run_output_path()}/artifacts', "
    # warn if data is potentially wrong
    if use_global_cache:
        cache_dir_arg = ""
        dry_warning = ""
        if not suppress_global_warning:
            output_lines.extend(
                [
                    "# %%",
                    "'''",
                    "<span style='color: orange;'><b>WARNING - </b></span>Experiment was not run with a `--store-full` flag, and so is simply using the project-wide cache rather than a specific experiment run cache. Any recent experiment runs since this notebook was created may have altered cached data."
                    "'''",
                    "",
                ]
            )
    else:
        dry_warning = "# Note that if this experiment uses lazy artifacts, you will want to remove the `dry=True` args below"

    if errored:
        output_lines.extend(
            [
                "# %%",
                "'''",
                "<span style='color: red;'><b>WARNING - </b></span>This experiment run did not complete due to an exception."
                "'''",
                "",
            ]
        )

    # imports and logger lines
    output_lines.extend(
        [
            "# %%",
            f"%cd {directory_change_back}",
            "",
            "# %%",
            "import logging",
            "from curifactory import ArtifactManager, experiment",
            "",
            "# %%",
            "logger = logging.getLogger()",
            "logger.setLevel(logging.INFO)",
            "",
            "# %%",
            dry_warning,
            f'manager = ArtifactManager("{experiment_name}", {cache_dir_arg} dry=True)',
            f'experiment.run_experiment("{experiment_name}", {str(param_files)}, dry=True, mngr=manager)',
            "",
            "# %%",
        ]
    )

    for i, param in enumerate(manager.records):
        output_lines.extend(
            [
                f"records{i} = manager.records[{i}]",
                f"state{i} = manager.records[{i}].state",
            ]
        )

        if manager.records[i].params is not None:
            output_lines.extend(
                [
                    f'print("state{i} - (" + records{i}.params.name + ") stages: " + str(records{i}.stages))',
                    f'print("keys: " + str(state{i}.keys()) + "\\n")',
                ]
            )
        else:
            output_lines.extend(
                [
                    f'print("state{i} - ((aggregate record)) stages: " + str(records{i}.stages))',
                    f'print("keys: " + str(state{i}.keys()) + "\\n")',
                ]
            )
        output_lines.append("")

    script_path = path + ".py"
    notebook_path = path + ".ipynb"

    output_lines = [line + "\n" for line in output_lines]

    with open(script_path, "w") as outfile:
        outfile.writelines(output_lines)

    # run ipynb-py-convert
    logging.info("Converting...")
    # utils.get_command_output(["ipynb-py-convert", script_path, notebook_path])

    cmd_array = ["ipynb-py-convert", script_path, notebook_path]
    print(*cmd_array)
    with subprocess.Popen(
        cmd_array, stderr=subprocess.PIPE, stdout=subprocess.PIPE, bufsize=1, text=True
    ) as p:
        for line in p.stdout:
            print(line, end="")  # process line here
        for line in p.stderr:
            print(line, end="")  # process line here
    logging.info("Output experiment notebook at %s", notebook_path)
    os.remove(script_path)


def regex_lister(module_name, regex, try_import=True):
    """Used by both list_experiments and list_params. This scans every file in the passed
    folder for the requested regex, and tries to import the files that have a match."""

    names = []

    path = module_name.replace(".", "/")

    if not os.path.exists(path):
        print(
            f"\t[WARNING - path: '{path}' does not exist. Double check curifactory_config.json]"
        )
        return []

    files = []
    # find all the python filenames in all subdirectories of the module
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith(".py"):
                files.append(os.path.join(dirpath, filename))

    for filename in files:
        with open(filename) as infile:
            lines = infile.readlines()
            # check for a run() method
            results = [re.findall(regex, line) for line in lines]

            # remove empty arrays from findall search
            clean_results = []
            for result in results:
                if result != []:
                    clean_results.extend(result)

            # did we find a run()/get_params()?
            if len(clean_results) > 0:
                non_pyextension_name = filename[:-3].replace("/", ".")
                # remove the module_name from the full filename
                non_pyextension_name = non_pyextension_name[len(module_name) + 1 :]

                if try_import:
                    # see if it's valid
                    try:
                        importlib.import_module(f"{module_name}.{non_pyextension_name}")
                        comment = utils.get_py_opening_comment(lines)
                        if comment != "":
                            names.append(
                                non_pyextension_name
                                + " - "
                                + utils.get_py_opening_comment(lines)
                            )
                        else:
                            names.append(non_pyextension_name)
                    except Exception as e:
                        names.append(non_pyextension_name + " [ERROR - " + str(e) + "]")
                else:
                    # we have a non-validity check option for speed, this is used
                    # for the argcomplete stuff
                    names.append(non_pyextension_name)

    return names


def list_experiments():
    """Print out all valid experiments that have a :code:`def run()` function, including
    any top-of-file docstrings associated with each."""
    config = utils.get_configuration()
    module = config["experiments_module_name"]
    experiment_names = regex_lister(module, r"^def run\(.*\)")
    experiment_names.sort()
    return experiment_names


def list_params():
    """Print out all valid parameter files that have a :code:`def get_params()`
    function, including any top-of-file docstrings associated with each."""
    config = utils.get_configuration()
    experiment_module = config["experiments_module_name"]
    params_module = config["params_module_name"]

    param_names = []
    param_names.extend(regex_lister(params_module, r"^def get_params\(.*\)"))
    param_names.extend(regex_lister(experiment_module, r"^def get_params\(.*\)"))
    param_names.sort()
    return param_names


def experiments_completer(**kwargs) -> list[str]:
    # argcomplete experiment completer
    config = utils.get_configuration()

    experiments_path = config["experiments_module_name"].replace(".", "/")
    files = (
        subprocess.run(
            f"cd {experiments_path} && grep -rl '^def run('",
            shell=True,
            capture_output=True,
        )
        .stdout.decode("utf-8")[:-1]
        .split("\n")
    )
    files = [file[:-3] for file in files if file != ""]
    files.sort()
    return [filename.replace("/", ".") for filename in files]


def params_completer(**kwargs) -> list[str]:
    # argcomplete -p completer
    config = utils.get_configuration()
    experiments_path = config["experiments_module_name"].replace(".", "/")
    params_path = config["params_module_name"].replace(".", "/")

    experiment_files = (
        subprocess.run(
            f"cd {experiments_path} && grep -rl '^def get_params('",
            shell=True,
            capture_output=True,
        )
        .stdout.decode("utf-8")[:-1]
        .split("\n")
    )
    experiment_files = [file[:-3] for file in experiment_files if file != ""]

    param_files = (
        subprocess.run(
            f"cd {params_path} && grep -rl '^def get_params('",
            shell=True,
            capture_output=True,
        )
        .stdout.decode("utf-8")[:-1]
        .split("\n")
    )
    param_files = [file[:-3] for file in param_files if file != ""]
    files = param_files + experiment_files
    files.sort()
    return [filename.replace("/", ".") for filename in files]


def main():
    """'Main' command line entrypoint, parses command line flags and makes the
    appropriate :code:`run_experiment()` call."""
    parser = argparse.ArgumentParser(
        description="Run a given curifactory experiment with specified parameters.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
    experiment my_experiment -p some_params1 -p some_params2
    experiment my_experiment_with_params
    experiment my_experiment_with_params --overwrite

    experiment ls  # lists all available experiments and parameters
    experiment reports --port 8000 --host 0.0.0.0
""",
    )
    parser.add_argument("experiment_name").completer = experiments_completer

    parser.add_argument(
        "--notes",
        nargs="?",  # this is how we allow it to be specified either as a flag or with a value
        dest="notes",
        const="",
        default=False,
        help="Associate some notes with this run. You can either directly specify a string to this flag, or leave it blank to open a notes file in your editor.",
    )
    parser.add_argument(
        "--no-dag",
        dest="no_dag",
        action="store_true",
        help="Specifying this flag disables DAG-mode execution and prevents the pre-execution mapping of the experiment. Instead, the experiment will be run straight through using only the cache to determine if a stage needs to execute or not.",
    )

    parameters_group = parser.add_argument_group(
        "Parameterization",
        "Choose which parameter files and which parameter sets to use.",
    )
    outputs_group = parser.add_argument_group(
        "Outputs", "Control what gets created from an experiment run."
    )
    caching_group = parser.add_argument_group(
        "Caching", "Configure cache usage (re-entrancy) and lazy artifacts."
    )
    display_group = parser.add_argument_group(
        "Display", "Configure console output during experiment execution."
    )
    parallel_group = parser.add_argument_group("Parallel execution")
    reports_group = parser.add_argument_group(
        "Reports", "Arguments for use with 'experiment reports'"
    )

    # ---- PARAMETERS ----
    parameters_group.add_argument(
        "-p",
        "--params",
        dest="parameters_name",
        action="append",
        help="The name of a parameters python file. Does not need to include path or .py. You can specify multiple -p arguments to run all parameters",
    ).completer = params_completer
    parameters_group.add_argument(
        "-n",
        "--names",
        dest="param_set_names",
        action="append",
        help="The name of a specific parameter set within one of the specified parameters files. Using this will run only the specified parameter set(s).",
    )
    parameters_group.add_argument(
        "--indices",
        dest="param_set_indices",
        action="append",
        help="A single or range of indices of parameter sets to use within the specified parameters files. Note that specifying this will run those indices from all specified parameters files. Specify ranges like '1-5'.",
    )
    parameters_group.add_argument(
        "--global-indices",
        dest="global_indices",
        action="append",
        help="A single or range of indices of argsets to use out of the total list of argsets passed to the experiment.",
    )

    # ---- OUTPUTS ----
    outputs_group.add_argument(
        "--notebook",
        dest="notebook",
        action="store_true",
        help="Generates a jupyter notebook to explore the results.",
    )
    outputs_group.add_argument(
        "--docker",
        dest="docker",
        action="store_true",
        help="Generates a docker image with the used cached files and a current copy of the codebase.",
    )
    outputs_group.add_argument(
        "--no-log",
        dest="no_log",
        action="store_true",
        help="Specify this flag to not store the log.",
    )
    outputs_group.add_argument(
        "--dry",
        dest="dry",
        action="store_true",
        help="Do a dry run: suppresses writing any cached objects and modifying stores.",
    )
    outputs_group.add_argument(
        "--log-errors",
        dest="log_errors",
        action="store_true",
        help="Include errors and stack traces in output logs. NOTE: this redirects stderr - output from libraries like tqdm will be logged as well.",
    )

    # ---- CACHING ----
    caching_group.add_argument(
        "-c",
        "--cache",
        dest="cache_dir",
        help="Specify a different directory to use as the cache.",
    )
    caching_group.add_argument(
        "-s",
        "--store-full",
        dest="store_full",
        action="store_true",
        help="Store a copy of environment information and every cached object in a folder named with the experiment reference.",
    )
    caching_group.add_argument(
        "--overwrite",
        dest="overwrite_override",
        action="store_true",
        help="Overwrite current cached objects",
    )
    caching_group.add_argument(
        "--overwrite-stage",
        dest="overwrite_stages",
        action="append",
        help="Names of specific stages to overwrite",
    )
    caching_group.add_argument(
        "--dry-cache",
        dest="dry_cache",
        action="store_true",
        help="Do a dry cache run: this still modifies stores and runs reports but does not write anything into the cache. This is recommended when running an experiment with a cache directory from --store-full.",
    )
    caching_group.add_argument(
        "--prefix",
        dest="prefix",
        default=None,
        help="Specify a custom prefix to use for caching rather than the experiment name. This can be useful if multiple similar experiments can use the same cached objects. Note that this does not change the reference name.",
    )
    caching_group.add_argument(
        "--lazy",
        dest="lazy",
        action="store_true",
        help="Attempts to lazy-cache all outputs from all stages. WARNING: if cachers are not specified throughout the stages, PickleCachers will be inserted. As pickle serialization does not work on some complex objects, this may fail. Specify valid cachers for all stage outputs to mitigate this.",
    )
    caching_group.add_argument(
        "--ignore-lazy",
        dest="ignore_lazy",
        action="store_true",
        help="Run the experiment without any lazy object caching, this can save time when memory is less of an issue.",
    )

    # ---- DISPLAY ----
    display_group.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Log at debug level.",
    )
    display_group.add_argument(
        "--map",
        dest="map_only",
        action="store_true",
        help="Specifying this _only_ runs the pre-execution record and stage mapping for the experiment, and prints out the resulting DAG information before immediately exiting. Specifying this implies --dry. You can use this flag to check which artifacts are found in cache.",
    )
    display_group.add_argument(
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Suppress all log output to console.",
    )
    display_group.add_argument(
        "--no-color", dest="no_color", action="store_true", help="Less fancy colors."
    )
    display_group.add_argument(
        "--progress",
        dest="progress",
        action="store_true",
        help="Display fancy progress bars! Note that this uses rich live and can break existing tqdm progress bars and other things.",
    )
    display_group.add_argument(
        "--plain",
        dest="plain",
        action="store_true",
        help="Print normal logging rather than rich colored logs. This will output the exact same text printed into the file log.",
    )

    # ---- PARALLEL ----
    parallel_group.add_argument(
        "--parallel",
        dest="parallel",
        default=None,
        help="Divide the parameters among n subprocesses in order to cache intermediate results in parallel, and then re-run in original process with cached data.",
    )
    parallel_group.add_argument(
        "--parallel-safe",
        dest="parallel_mode",
        action="store_true",
        help="Runs an experiment and caches intermediate values but does not report or touch experiment store and parameter registry. Note that using this flag manually is not equivalent to running an experiment with the --parallel flag, as file locks don't get used when calling this manually, potentially creating race conditions.",
    )

    # ---- REPORTS ----
    reports_group.add_argument(
        "--port",
        dest="port",
        default=8080,
        help="Only used for 'experiment reports', specifies which port to run the simple server on.",
    )
    reports_group.add_argument(
        "--host",
        dest="host",
        default="127.0.0.1",
        help="Only used for 'experiment reports', specifies which hostname to run the simple server on.",
    )
    reports_group.add_argument(
        "--update",
        dest="update",
        action="store_true",
        help="Only used for 'experiment reports', updates the report index with all exisiting reports in the reports file. This is to handle if you pull in reports from other machines.",
    )
    argcomplete.autocomplete(parser, always_complete_options=False)

    # fix any missing quotes in run line
    command_parts = sys.argv[1:]

    fixed_parts = []
    for index, part in enumerate(command_parts):
        if " " in part and not part.startswith('"') and not part.endswith('"'):
            part = f'"{part}"'

            # for now we don't include the notes param, because the notes are already included in
            # the report render.
            if index > 0 and command_parts[index - 1] == "--notes":
                continue
        if part == "--notes":
            continue
        fixed_parts.append(part)

    run_string = "experiment " + " ".join(fixed_parts)

    args = parser.parse_args()
    params_list = args.parameters_name

    if args.experiment_name == "ls":
        sys.path.append(os.getcwd())
        print("EXPERIMENTS:")
        for experiment in list_experiments():
            print("\t" + experiment)
        print("\nPARAMS:")
        for param in list_params():
            print("\t" + param)
        return

    elif args.experiment_name == "reports":
        if args.update:
            config = utils.get_configuration()
            utils.init_logging(None)
            reporting.update_report_index(
                config["experiments_module_name"], config["reports_path"]
            )
            exit()
        os.chdir("reports")
        utils.run_command(
            ["python", "-m", "http.server", str(args.port), "--bind", args.host]
        )
        os.chdir("..")
        return

    # TODO: verify names exist
    # TODO: ensure folders exist (logs)

    log = True
    if args.no_log:
        log = False

    parallel = None
    if args.parallel is not None:
        parallel = int(args.parallel)

    if args.notes != "" and not args.notes:
        args.notes = None

    run_experiment(
        args.experiment_name,
        params_list,
        args.overwrite_override,
        args.cache_dir,
        log=log,
        log_debug=args.verbose,
        dry=args.dry,
        dry_cache=args.dry_cache,
        store_full=args.store_full,
        log_errors=args.log_errors,
        prefix=args.prefix,
        build_docker=args.docker,
        build_notebook=args.notebook,
        run_string=run_string,
        stage_overwrites=args.overwrite_stages,
        param_set_names=args.param_set_names,
        param_set_indices=args.param_set_indices,
        global_param_set_indices=args.global_indices,
        parallel=parallel,
        parallel_mode=args.parallel_mode,
        lazy=args.lazy,
        ignore_lazy=args.ignore_lazy,
        no_dag=args.no_dag,
        map_only=args.map_only,
        no_color=args.no_color,
        quiet=args.quiet,
        progress=args.progress,
        plain=args.plain,
        notes=args.notes,
    )


if __name__ == "__main__":
    main()



"""Convenience class for grouping stages and automatically passing a record between them."""

from curifactory.manager import ArtifactManager, Record


class Procedure:
    """A defined list of stages to run in sequence, creating a record to associate
    with them.

    For the stage list, specify only the names of the functions.

    Example:

        .. code-block:: python

            @stage(...)
            def data_stage(...):
                # ...

            @stage(...)
            def model_stage(...):
                # ...

            proc = Procedure(
                [
                    data_stage,
                    model_stage
                ],
                mngr)

    Args:
        stages: A list of function names that are wrapped in ``@stage`` or ``@aggregate``
            decorators. Note that if using an aggregate state, it _must_ be the first one in the
            list.
        manager (ArtifactManager): The manager to associate this procedure and corresponding record
            with. If you specify :code:`None`, one will be created with the default constructor.
        name (str): An optional name for the procedure. Currently unused, may eventually be put into
            logging or reporting.
        previous_proc (Procedure): If specified and this procedure begins with an aggregate stage,
            use the ``previous_proc.records`` list of records.
        records (List[Record]): If specified and this procedure begins with an aggregate stage, use
            this list of records. Note that ``previous_proc`` takes precedence over this argument.

    Note:
        If a procedure begins with an aggregate stage and neither ``previous_proc`` nor
        ``records`` are specified, it will automatically grab all existing records from the
        artifact manager.
    """

    def __init__(
        self, stages, manager=None, name=None, previous_proc=None, records=None
    ):
        # NOTE: pass in previous procedure to auto aggregate across just the records
        # from that previous procedure
        self.name = name
        self.stages = stages
        self.record = None
        self.manager = manager
        if manager is None:
            self.manager = (
                ArtifactManager()
            )  # TODO: notably this doesn't respect the config.

        self.records = []  # keeps track of all records run through this procedure
        # this is just for convenience for being able to aggregate off previous proc
        self.previous_proc = previous_proc
        self.use_records = records

    def run(self, param_set, record=None, hide=False):
        """Run this procedure with the passed parameter set. This allows easily running
        multiple parameter sets through the same set of stages and automatically getting a separate
        record for each.

        Args:
            param_set (ExperimentParameters): The parameteter set to put into the record created for
                this procedure.
            record (Record): If you have a specific record you want the procedure to use (e.g.
                if you're chaining multiple procedures and already have an applicable record to
                use from the previous one), pass it here. If unspecified, a new record will
                automatically be created for the passed args and relevant artifact manager.
            hide (bool): If ``True``, don't add the created record to the artifact manager.

        Returns:
            The returned output from the last stage in ``self.stages``.
        """
        # create a new record as needed
        if record is None:
            self.record = Record(self.manager, param_set, hide=hide)
            self.records.append(self.record)
        else:
            self.record = record

        for index, stage in enumerate(self.stages):
            if index == 0 and self.previous_proc is not None:
                # the first stage of a proc based on previous proc needs to be
                # an aggregate proc
                # if you don't explicitly pass previous proc, it will take all
                # records from manager
                output = stage(self.record, self.previous_proc.records)
            elif index == 0 and self.use_records is not None:
                output = stage(self.record, self.use_records)
            else:
                # Note that if the previous two conditions _weren't_ applicable but
                # the first stage is still an aggregate, the aggregate decorator correctly
                # handles converting the ``None`` that will be passed in for the records,
                # to the full list of records on the manager.
                output = stage(self.record)  # have to pass record into each stage

        return output
